# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, division
from functools import reduce  # TODO PY2 compatibility
import operator
import itertools


import numpy as np


from zarr.compressors import get_compressor_cls
from zarr.util import is_total_slice, normalize_array_selection, \
    get_chunk_range, human_readable_size, normalize_resize_args
from zarr.meta import decode_metadata, encode_metadata
from zarr.attrs import Attributes
from zarr.compat import itervalues
from zarr.errors import ReadOnlyError


class Array(object):
    """Instantiate an array from an initialised store.

    Parameters
    ----------
    store : MutableMapping
        Array store, already initialised.
    readonly : bool, optional
        True if array should be protected against modification.

    Attributes
    ----------
    store
    readonly
    shape
    chunks
    dtype
    compression
    compression_opts
    fill_value
    order
    attrs
    size
    itemsize
    nbytes
    nbytes_stored
    initialized
    cdata_shape

    Methods
    -------
    __getitem__
    __setitem__
    resize
    append

    Examples
    --------
    >>> import zarr
    >>> store = dict()
    >>> zarr.init_store(store, shape=(10000, 10000), chunks=(1000, 1000))
    >>> z = zarr.Array(store)
    >>> z
    zarr.core.Array((10000, 10000), float64, chunks=(1000, 1000), order=C)
      compression: blosc; compression_opts: {'clevel': 5, 'cname': 'blosclz', 'shuffle': 1}
      nbytes: 762.9M; nbytes_stored: 320; ratio: 2500000.0; initialized: 0/100
      store: builtins.dict

    """  # flake8: noqa

    def __init__(self, store, readonly=False):
        # N.B., expect at this point store is fully initialised with all
        # configuration metadata fully specified and normalised

        #: store docstring
        self._store = store  #: inline docstring
        self._readonly = readonly

        # initialise metadata
        try:
            meta_bytes = store['meta']
        except KeyError:
            raise ValueError('store has no metadata')
        else:
            meta = decode_metadata(meta_bytes)
            self._meta = meta
            self._shape = meta['shape']
            self._chunks = meta['chunks']
            self._dtype = meta['dtype']
            self._compression = meta['compression']
            self._compression_opts = meta['compression_opts']
            self._fill_value = meta['fill_value']
            self._order = meta['order']
            compressor_cls = get_compressor_cls(self._compression)
            self._compressor = compressor_cls(self._compression_opts)

        # initialise attributes
        self._attrs = Attributes(store, readonly=readonly)

    def flush_metadata(self):
        meta = dict(shape=self._shape, chunks=self._chunks, dtype=self._dtype,
                    compression=self._compression,
                    compression_opts=self._compression_opts,
                    fill_value=self._fill_value, order=self._order)
        self._store['meta'] = encode_metadata(meta)

    @property
    def store(self):
        """A MutableMapping providing the underlying storage for the array."""
        return self._store

    @property
    def readonly(self):
        """A boolean, True if write operations are not permitted."""
        return self._readonly

    @property
    def shape(self):
        """A tuple of integers describing the length of each dimension of
        the array."""
        return self._shape

    @property
    def chunks(self):
        """A tuple of integers describing the length of each dimension of a
        chunk of the array."""
        return self._chunks

    @property
    def dtype(self):
        """The NumPy data type."""
        return self._dtype

    @property
    def compression(self):
        """A string naming the primary compression algorithm used to
        compress chunks of the array."""
        return self._compression

    @property
    def compression_opts(self):
        """Parameters controlling the behaviour of the primary compression
        algorithm."""
        return self._compression_opts

    @property
    def fill_value(self):
        """A value used for uninitialized portions of the array."""
        return self._fill_value

    @property
    def order(self):
        """A string indicating the order in which bytes are arranged within
        chunks of the array."""
        return self._order

    @property
    def attrs(self):
        """A MutableMapping containing user-defined attributes. Note that
        attribute values must be JSON serializable."""
        return self._attrs

    @property
    def size(self):
        """The total number of elements in the array."""
        return reduce(operator.mul, self._shape)

    @property
    def itemsize(self):
        """The size in bytes of each item in the array."""
        return self._dtype.itemsize

    @property
    def nbytes(self):
        """The total number of bytes that would be required to store the
        array without compression."""
        return self.size * self.itemsize

    @property
    def nbytes_stored(self):
        """The total number of stored bytes of data for the array. This
        includes storage required for configuration metadata and user
        attributes encoded as JSON."""
        if hasattr(self._store, 'size'):
            # pass through
            return self._store.size
        elif isinstance(self._store, dict):
            # cheap to compute by summing length of values
            return sum(len(v) for v in itervalues(self._store))
        else:
            return -1

    @property
    def initialized(self):
        """The number of chunks that have been initialized with some data."""
        # N.B., expect 'meta' and 'attrs' keys in store also, so subtract 2
        return len(self._store) - 2

    @property
    def cdata_shape(self):
        """A tuple of integers describing the number of chunks along each
        dimension of the array."""
        return tuple(
            int(np.ceil(s / c)) for s, c in zip(self._shape, self._chunks)
        )

    def __array__(self):
        return self[:]

    def __len__(self):
        return self.shape[0]

    def __getitem__(self, item):
        """Retrieve data for some portion of the array. Most NumPy-style
        slicing operations are supported.

        Returns
        -------
        out : ndarray
            A NumPy array containing the data for the requested region.

        Examples
        --------

        Setup a 1-dimensional array::

            >>> import zarr
            >>> import numpy as np
            >>> z = zarr.array(np.arange(100000000), chunks=1000000, dtype='i4')
            >>> z
            zarr.core.Array((100000000,), int32, chunks=(1000000,), order=C)
              compression: blosc; compression_opts: {'clevel': 5, 'cname': 'blosclz', 'shuffle': 1}
              nbytes: 381.5M; nbytes_stored: 6.8M; ratio: 56.0; initialized: 100/100
              store: builtins.dict

        Take some slices::

            >>> z[5]
            5
            >>> z[:5]
            array([0, 1, 2, 3, 4], dtype=int32)
            >>> z[-5:]
            array([99999995, 99999996, 99999997, 99999998, 99999999], dtype=int32)
            >>> z[5:10]
            array([5, 6, 7, 8, 9], dtype=int32)
            >>> z[:]
            array([       0,        1,        2, ..., 99999997, 99999998, 99999999], dtype=int32)

        Setup a 2-dimensional array::

            >>> import zarr
            >>> import numpy as np
            >>> z = zarr.array(np.arange(100000000).reshape(10000, 10000),
            ...                chunks=(1000, 1000), dtype='i4')
            >>> z
            zarr.core.Array((10000, 10000), int32, chunks=(1000, 1000), order=C)
              compression: blosc; compression_opts: {'clevel': 5, 'cname': 'blosclz', 'shuffle': 1}
              nbytes: 381.5M; nbytes_stored: 10.0M; ratio: 38.0; initialized: 100/100
              store: builtins.dict

        Take some slices::

            >>> z[2, 2]
            20002
            >>> z[:2, :2]
            array([[    0,     1],
                   [10000, 10001]], dtype=int32)
            >>> z[:2]
            array([[    0,     1,     2, ...,  9997,  9998,  9999],
                   [10000, 10001, 10002, ..., 19997, 19998, 19999]], dtype=int32)
            >>> z[:, :2]
            array([[       0,        1],
                   [   10000,    10001],
                   [   20000,    20001],
                   ...,
                   [99970000, 99970001],
                   [99980000, 99980001],
                   [99990000, 99990001]], dtype=int32)
            >>> z[:]
            array([[       0,        1,        2, ...,     9997,     9998,     9999],
                   [   10000,    10001,    10002, ...,    19997,    19998,    19999],
                   [   20000,    20001,    20002, ...,    29997,    29998,    29999],
                   ...,
                   [99970000, 99970001, 99970002, ..., 99979997, 99979998, 99979999],
                   [99980000, 99980001, 99980002, ..., 99989997, 99989998, 99989999],
                   [99990000, 99990001, 99990002, ..., 99999997, 99999998, 99999999]], dtype=int32)

        """  # flake8: noqa

        # normalize selection
        selection = normalize_array_selection(item, self._shape)

        # determine output array shape
        out_shape = tuple(s.stop - s.start for s in selection
                          if isinstance(s, slice))

        # setup output array
        out = np.empty(out_shape, dtype=self._dtype, order=self._order)

        # determine indices of chunks overlapping the selection
        chunk_range = get_chunk_range(selection, self._chunks)

        # iterate over chunks in range
        for cidx in itertools.product(*chunk_range):

            # determine chunk offset
            offset = [i * c for i, c in zip(cidx, self._chunks)]

            # determine region within output array
            out_selection = tuple(
                slice(max(0, o - s.start),
                      min(o + c - s.start, s.stop - s.start))
                for s, o, c, in zip(selection, offset, self._chunks)
                if isinstance(s, slice)
            )

            # determine region within chunk
            chunk_selection = tuple(
                slice(max(0, s.start - o), min(c, s.stop - o))
                if isinstance(s, slice)
                else s - o
                for s, o, c in zip(selection, offset, self._chunks)
            )

            # obtain the destination array as a view of the output array
            if out_selection:
                dest = out[out_selection]
            else:
                dest = out

            # load chunk selection into output array
            self._chunk_getitem(cidx, chunk_selection, dest)

        if out.shape:
            return out
        else:
            return out[()]

    def __setitem__(self, key, value):
        """Modify data for some portion of the array.

        Examples
        --------

        Setup a 1-dimensional array::

            >>> import zarr
            >>> z = zarr.zeros(100000000, chunks=1000000, dtype='i4')
            >>> z
            zarr.core.Array((100000000,), int32, chunks=(1000000,), order=C)
              compression: blosc; compression_opts: {'clevel': 5, 'cname': 'blosclz', 'shuffle': 1}
              nbytes: 381.5M; nbytes_stored: 295; ratio: 1355932.2; initialized: 0/100
              store: builtins.dict

        Set all array elements to the same scalar value::

            >>> z[:] = 42
            >>> z[:]
            array([42, 42, 42, ..., 42, 42, 42], dtype=int32)

        Set a portion of the array::

            >>> z[:100] = np.arange(100)
            >>> z[-100:] = np.arange(100)[::-1]
            >>> z[:]
            array([0, 1, 2, ..., 2, 1, 0], dtype=int32)

        Setup a 2-dimensional array::

            >>> z = zarr.zeros((10000, 10000), chunks=(1000, 1000), dtype='i4')
            >>> z
            zarr.core.Array((10000, 10000), int32, chunks=(1000, 1000), order=C)
              compression: blosc; compression_opts: {'clevel': 5, 'cname': 'blosclz', 'shuffle': 1}
              nbytes: 381.5M; nbytes_stored: 317; ratio: 1261829.7; initialized: 0/100
              store: builtins.dict

        Set all array elements to the same scalar value::

            >>> z[:] = 42
            >>> z[:]
            array([[42, 42, 42, ..., 42, 42, 42],
                   [42, 42, 42, ..., 42, 42, 42],
                   [42, 42, 42, ..., 42, 42, 42],
                   ...,
                   [42, 42, 42, ..., 42, 42, 42],
                   [42, 42, 42, ..., 42, 42, 42],
                   [42, 42, 42, ..., 42, 42, 42]], dtype=int32)

        Set a portion of the array::

            >>> z[0, :] = np.arange(z.shape[1])
            >>> z[:, 0] = np.arange(z.shape[0])
            >>> z[:]
            array([[   0,    1,    2, ..., 9997, 9998, 9999],
                   [   1,   42,   42, ...,   42,   42,   42],
                   [   2,   42,   42, ...,   42,   42,   42],
                   ...,
                   [9997,   42,   42, ...,   42,   42,   42],
                   [9998,   42,   42, ...,   42,   42,   42],
                   [9999,   42,   42, ...,   42,   42,   42]], dtype=int32)

        """

        # guard conditions
        if self._readonly:
            raise ReadOnlyError('array is read-only')

        # normalize selection
        selection = normalize_array_selection(key, self._shape)

        # check value shape
        expected_shape = tuple(
            s.stop - s.start for s in selection
            if isinstance(s, slice)
        )
        if np.isscalar(value):
            pass
        elif expected_shape != value.shape:
            raise ValueError('value has wrong shape, expecting %s, found %s'
                             % (str(expected_shape),
                                str(value.shape)))

        # determine indices of chunks overlapping the selection
        chunk_range = get_chunk_range(selection, self._chunks)

        # iterate over chunks in range
        for cidx in itertools.product(*chunk_range):

            # determine chunk offset
            offset = [i * c for i, c in zip(cidx, self._chunks)]

            # determine required index range within chunk
            chunk_selection = tuple(
                slice(max(0, s.start - o), min(c, s.stop - o))
                if isinstance(s, slice)
                else s - o
                for s, o, c in zip(selection, offset, self._chunks)
            )

            if np.isscalar(value):

                # put data
                self._chunk_setitem(cidx, chunk_selection, value)

            else:
                # assume value is array-like

                # determine index within value
                value_selection = tuple(
                    slice(max(0, o - s.start),
                          min(o + c - s.start, s.stop - s.start))
                    for s, o, c in zip(selection, offset, self._chunks)
                    if isinstance(s, slice)
                )

                # put data
                self._chunk_setitem(cidx, chunk_selection,
                                    value[value_selection])

    def _chunk_getitem(self, cidx, item, dest):
        """Obtain part or whole of a chunk.

        Parameters
        ----------
        cidx : tuple of ints
            Indices of the chunk.
        item : tuple of slices
            Location of region within the chunk.
        dest : ndarray
            Numpy array to store result in.

        """

        try:

            # obtain compressed data for chunk
            ckey = '.'.join(map(str, cidx))
            cdata = self._store[ckey]

        except KeyError:

            # chunk not initialized
            if self._fill_value is not None:
                dest.fill(self._fill_value)

        else:

            if is_total_slice(item, self._chunks) and \
                    ((self._order == 'C' and dest.flags.c_contiguous) or
                     (self._order == 'F' and dest.flags.f_contiguous)):

                # optimisation: we want the whole chunk, and the destination is
                # contiguous, so we can decompress directly from the chunk
                # into the destination array
                self._compressor.decompress(cdata, dest)

            else:

                # decompress chunk
                chunk = np.empty(self._chunks, dtype=self._dtype,
                                 order=self._order)
                self._compressor.decompress(cdata, chunk)

                # set data in output array
                # (split into two lines for profiling)
                tmp = chunk[item]
                if dest.shape:
                    dest[:] = tmp
                else:
                    dest[()] = tmp

    def _chunk_setitem(self, cidx, key, value):
        """Replace part or whole of a chunk.

        Parameters
        ----------
        cidx : tuple of ints
            Indices of the chunk.
        key : tuple of slices
            Location of region within the chunk.
        value : scalar or ndarray
            Value to set.

        """

        if is_total_slice(key, self._chunks):

            # optimisation: we are completely replacing the chunk, so no need
            # to access the existing chunk data

            if np.isscalar(value):

                # setup array filled with value
                chunk = np.empty(self._chunks, dtype=self._dtype,
                                 order=self._order)
                chunk.fill(value)

            else:

                # ensure array is contiguous
                if self._order == 'F':
                    chunk = np.asfortranarray(value, dtype=self._dtype)
                else:
                    chunk = np.ascontiguousarray(value, dtype=self._dtype)

        else:
            # partially replace the contents of this chunk

            try:

                # obtain compressed data for chunk
                ckey = '.'.join(map(str, cidx))
                cdata = self._store[ckey]

            except KeyError:

                # chunk not initialized
                chunk = np.empty(self._chunks, dtype=self._dtype,
                                 order=self._order)
                if self._fill_value is not None:
                    chunk.fill(self._fill_value)

            else:

                # decompress chunk
                chunk = np.empty(self._chunks, dtype=self._dtype,
                                 order=self._order)
                self._compressor.decompress(cdata, chunk)

            # modify
            chunk[key] = value

        # compress
        cdata = self._compressor.compress(chunk)

        # store
        ckey = '.'.join(map(str, cidx))
        self._store[ckey] = cdata

    def __repr__(self):
        r = '%s.%s(' % (type(self).__module__, type(self).__name__)
        r += '%s' % str(self._shape)
        r += ', %s' % str(self._dtype)
        r += ', chunks=%s' % str(self._chunks)
        r += ', order=%s' % self._order
        r += ')'
        r += '\n  compression: %s' % self._compression
        r += '; compression_opts: %s' % str(self._compression_opts)
        r += '\n  nbytes: %s' % human_readable_size(self.nbytes)
        if self.nbytes_stored > 0:
            r += '; nbytes_stored: %s' % human_readable_size(
                self.nbytes_stored)
            r += '; ratio: %.1f' % (self.nbytes / self.nbytes_stored)
        n_chunks = reduce(operator.mul, self.cdata_shape)
        r += '; initialized: %s/%s' % (self.initialized, n_chunks)
        r += '\n  store: %s.%s' % (type(self._store).__module__,
                                   type(self._store).__name__)
        return r

    def __getstate__(self):
        return self._store, self._readonly

    def __setstate__(self, state):
        self.__init__(*state)

    def resize(self, *args):
        """Change the shape of the array by growing or shrinking one or more
        dimensions.

        Examples
        --------
        >>> import zarr
        >>> z = zarr.zeros(shape=(10000, 10000), chunks=(1000, 1000))
        >>> z
        zarr.core.Array((10000, 10000), float64, chunks=(1000, 1000), order=C)
          compression: blosc; compression_opts: {'clevel': 5, 'cname': 'blosclz', 'shuffle': 1}
          nbytes: 762.9M; nbytes_stored: 317; ratio: 2523659.3; initialized: 0/100
          store: builtins.dict
        >>> z.resize(20000, 10000)
        >>> z
        zarr.core.Array((20000, 10000), float64, chunks=(1000, 1000), order=C)
          compression: blosc; compression_opts: {'clevel': 5, 'cname': 'blosclz', 'shuffle': 1}
          nbytes: 1.5G; nbytes_stored: 317; ratio: 5047318.6; initialized: 0/200
          store: builtins.dict
        >>> z.resize(30000, 1000)
        >>> z
        zarr.core.Array((30000, 1000), float64, chunks=(1000, 1000), order=C)
          compression: blosc; compression_opts: {'clevel': 5, 'cname': 'blosclz', 'shuffle': 1}
          nbytes: 228.9M; nbytes_stored: 316; ratio: 759493.7; initialized: 0/30
          store: builtins.dict

        Notes
        -----
        When resizing an array, the data are not rearranged in any way.

        If one or more dimensions are shrunk, any chunks falling outside the
        new array shape will be deleted from the underlying store.

        """  # flake8: noqa

        # guard conditions
        if self._readonly:
            raise ReadOnlyError('array is read-only')

        # normalize new shape argument
        old_shape = self._shape
        new_shape = normalize_resize_args(old_shape, *args)

        # determine the new number and arrangement of chunks
        chunks = self._chunks
        new_cdata_shape = tuple(int(np.ceil(s / c))
                                for s, c in zip(new_shape, chunks))

        # remove any chunks not within range
        for key in list(self._store):
            if key not in ['meta', 'attrs']:
                cidx = map(int, key.split('.'))
                if all(i < c for i, c in zip(cidx, new_cdata_shape)):
                    pass  # keep the chunk
                else:
                    del self._store[key]

        # update metadata
        self._shape = new_shape
        self.flush_metadata()

    def append(self, data, axis=0):
        """Append `data` to `axis`.

        Parameters
        ----------
        data : array_like
            Data to be appended.
        axis : int
            Axis along which to append.

        Notes
        -----
        The size of all dimensions other than `axis` must match between this
        array and `data`.

        Examples
        --------
        >>> import numpy as np
        >>> import zarr
        >>> a = np.arange(10000000, dtype='i4').reshape(10000, 1000)
        >>> z = zarr.array(a, chunks=(1000, 100))
        >>> z
        zarr.core.Array((10000, 1000), int32, chunks=(1000, 100), order=C)
          compression: blosc; compression_opts: {'clevel': 5, 'cname': 'blosclz', 'shuffle': 1}
          nbytes: 38.1M; nbytes_stored: 2.0M; ratio: 19.3; initialized: 100/100
          store: builtins.dict
        >>> z.append(a)
        >>> z
        zarr.core.Array((20000, 1000), int32, chunks=(1000, 100), order=C)
          compression: blosc; compression_opts: {'clevel': 5, 'cname': 'blosclz', 'shuffle': 1}
          nbytes: 76.3M; nbytes_stored: 4.0M; ratio: 19.3; initialized: 200/200
          store: builtins.dict
        >>> z.append(np.vstack([a, a]), axis=1)
        >>> z
        zarr.core.Array((20000, 2000), int32, chunks=(1000, 100), order=C)
          compression: blosc; compression_opts: {'clevel': 5, 'cname': 'blosclz', 'shuffle': 1}
          nbytes: 152.6M; nbytes_stored: 7.9M; ratio: 19.3; initialized: 400/400
          store: builtins.dict

        """

        # guard conditions
        if self._readonly:
            raise ReadOnlyError('array is read-only')

        # ensure data is array-like
        if not hasattr(data, 'shape') or not hasattr(data, 'dtype'):
            data = np.asanyarray(data)

        # ensure shapes are compatible for non-append dimensions
        self_shape_preserved = tuple(s for i, s in enumerate(self._shape)
                                     if i != axis)
        data_shape_preserved = tuple(s for i, s in enumerate(data.shape)
                                     if i != axis)
        if self_shape_preserved != data_shape_preserved:
            raise ValueError('shapes not compatible')

        # remember old shape
        old_shape = self._shape

        # determine new shape
        new_shape = tuple(
            self._shape[i] if i != axis else self._shape[i] + data.shape[i]
            for i in range(len(self._shape))
        )

        # resize
        self.resize(new_shape)

        # store data
        # noinspection PyTypeChecker
        append_selection = tuple(
            slice(None) if i != axis else slice(old_shape[i], new_shape[i])
            for i in range(len(self._shape))
        )
        self[append_selection] = data



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, division
from collections import MutableMapping
import os
import tempfile
import json


import numpy as np


from zarr.util import normalize_shape, normalize_chunks, normalize_order
from zarr.compressors import get_compressor_cls
from zarr.meta import encode_metadata


def init_store(store, shape, chunks, dtype=None, compression='default',
               compression_opts=None, fill_value=None,
               order='C', overwrite=False):
    """Initialise an array store with the given configuration.

    Parameters
    ----------
    store : MutableMapping
        A mapping that supports string keys and byte sequence values.
    shape : int or tuple of ints
        Array shape.
    chunks : int or tuple of ints
        Chunk shape.
    dtype : string or dtype, optional
        NumPy dtype.
    compression : string, optional
        Name of primary compression library, e.g., 'blosc', 'zlib', 'bz2',
        'lzma'.
    compression_opts : object, optional
        Options to primary compressor. E.g., for blosc, provide a dictionary
        with keys 'cname', 'clevel' and 'shuffle'.
    fill_value : object
        Default value to use for uninitialised portions of the array.
    order : {'C', 'F'}, optional
        Memory layout to be used within each chunk.
    overwrite : bool, optional
        If True, erase all data in `store` prior to initialisation.

    Examples
    --------
    >>> import zarr
    >>> store = dict()
    >>> zarr.init_store(store, shape=(10000, 10000), chunks=(1000, 1000))
    >>> sorted(store.keys())
    ['attrs', 'meta']
    >>> print(str(store['meta'], 'ascii'))
    {
        "chunks": [
            1000,
            1000
        ],
        "compression": "blosc",
        "compression_opts": {
            "clevel": 5,
            "cname": "blosclz",
            "shuffle": 1
        },
        "dtype": "<f8",
        "fill_value": null,
        "order": "C",
        "shape": [
            10000,
            10000
        ],
        "zarr_format": 1
    }
    >>> print(str(store['attrs'], 'ascii'))
    {}

    Notes
    -----
    The initialisation process involves normalising all array metadata,
    encoding as JSON and storing under the 'meta' key. User attributes are also
    initialised and stored as JSON under the 'attrs' key.

    """

    # guard conditions
    empty = len(store) == 0
    if not empty and not overwrite:
        raise ValueError('store is not empty')

    # normalise metadata
    shape = normalize_shape(shape)
    chunks = normalize_chunks(chunks, shape)
    dtype = np.dtype(dtype)
    compressor_cls = get_compressor_cls(compression)
    compression = compressor_cls.canonical_name
    compression_opts = compressor_cls.normalize_opts(
        compression_opts
    )
    order = normalize_order(order)

    # delete any pre-existing items in store
    store.clear()

    # initialise metadata
    meta = dict(shape=shape, chunks=chunks, dtype=dtype,
                compression=compression, compression_opts=compression_opts,
                fill_value=fill_value, order=order)
    store['meta'] = encode_metadata(meta)

    # initialise attributes
    store['attrs'] = json.dumps(dict()).encode('ascii')


class DirectoryStore(MutableMapping):
    """Mutable Mapping interface to a directory. Keys must be strings,
    values must be bytes-like objects.

    Parameters
    ----------
    path : string
        Location of directory.

    Examples
    --------
    >>> import zarr
    >>> store = zarr.DirectoryStore('example.zarr')
    >>> zarr.init_store(store, shape=(10000, 10000), chunks=(1000, 1000),
    ...                 fill_value=0, overwrite=True)
    >>> import os
    >>> sorted(os.listdir('example.zarr'))
    ['attrs', 'meta']
    >>> print(open('example.zarr/meta').read())
    {
        "chunks": [
            1000,
            1000
        ],
        "compression": "blosc",
        "compression_opts": {
            "clevel": 5,
            "cname": "blosclz",
            "shuffle": 1
        },
        "dtype": "<f8",
        "fill_value": 0,
        "order": "C",
        "shape": [
            10000,
            10000
        ],
        "zarr_format": 1
    }
    >>> print(open('example.zarr/attrs').read())
    {}
    >>> z = zarr.Array(store)
    >>> z
    zarr.core.Array((10000, 10000), float64, chunks=(1000, 1000), order=C)
      compression: blosc; compression_opts: {'clevel': 5, 'cname': 'blosclz', 'shuffle': 1}
      nbytes: 762.9M; nbytes_stored: 317; ratio: 2523659.3; initialized: 0/100
      store: zarr.storage.DirectoryStore
    >>> z[:] = 1
    >>> len(os.listdir('example.zarr'))
    102
    >>> sorted(os.listdir('example.zarr'))[0:5]
    ['0.0', '0.1', '0.2', '0.3', '0.4']
    >>> print(open('example.zarr/0.0', 'rb').read(10))
    b'\\x02\\x01\\x01\\x08\\x00\\x12z\\x00\\x00\\x80'

    See Also
    --------
    zarr.creation.open

    """  # flake8: noqa

    def __init__(self, path):

        # guard conditions
        path = os.path.abspath(path)
        if not os.path.exists(path):
            raise ValueError('path does not exist')
        elif not os.path.isdir(path):
            raise ValueError('path is not a directory')

        self.path = path

    def __getitem__(self, key):

        # guard conditions
        if key not in self:
            raise KeyError(key)

        with open(os.path.join(self.path, key), 'rb') as f:
            return f.read()

    def __setitem__(self, key, value):
        # accept any value that can be written to a file

        # destination path for key
        dest_path = os.path.join(self.path, key)

        # write to temporary file
        with tempfile.NamedTemporaryFile(mode='wb', delete=False,
                                         dir=self.path,
                                         prefix=key + '.',
                                         suffix='.partial') as f:
            f.write(value)
            temp_path = f.name

        # move temporary file into place
        if os.path.exists(dest_path):
            os.remove(dest_path)
        os.rename(temp_path, dest_path)

    def __delitem__(self, key):

        # guard conditions
        if key not in self:
            raise KeyError(key)

        os.remove(os.path.join(self.path, key))

    def __contains__(self, key):
        return os.path.isfile(os.path.join(self.path, key))

    def keys(self):
        for key in os.listdir(self.path):
            if os.path.isfile(os.path.join(self.path, key)):
                yield key

    def __iter__(self):
        return self.keys()

    def __len__(self):
        return sum(1 for _ in self.keys())

    @property
    def size(self):
        """Total size of all values in number of bytes."""
        return sum(os.path.getsize(os.path.join(self.path, key))
                   for key in self.keys())



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, division
import unittest
import array


import numpy as np
from numpy.testing import assert_array_equal
from nose.tools import eq_ as eq, assert_raises


from zarr.compressors import get_compressor_cls


class CompressorTests(object):

    compression = None

    @classmethod
    def init_compressor(cls, compression_opts=None):
        compression_cls = get_compressor_cls(cls.compression)
        compression_opts = compression_cls.normalize_opts(compression_opts)
        compressor = compression_cls(compression_opts)
        return compressor

    def _test_compress_decompress(self, compression_opts=None):

        comp = self.init_compressor(compression_opts)
        a = np.arange(1000, dtype='i4')
        cdata = comp.compress(a)
        assert isinstance(cdata, (bytes, bytearray, array.array))
        assert len(cdata) <= a.nbytes

        b = np.empty_like(a)
        comp.decompress(cdata, b)
        assert_array_equal(a, b)

    def test_compress_decompress_default(self):
        self._test_compress_decompress()


try:
    from zarr import blosc  # flake8: noqa
except ImportError:
    print('Blosc not available, skipping Blosc compressor tests')
else:

    class TestBloscCompressor(unittest.TestCase, CompressorTests):

        compression = 'blosc'

        def test_normalize_opts(self):
            cls = get_compressor_cls(self.compression)

            # test defaults
            opts = cls.normalize_opts(None)
            default_opts = dict(
                cname=cls.default_cname,
                clevel=cls.default_clevel,
                shuffle=cls.default_shuffle
            )
            eq(default_opts, opts)

            # test invalid args
            with assert_raises(ValueError):
                cls.normalize_opts(dict(cname='foo'))
            with assert_raises(ValueError):
                cls.normalize_opts(dict(clevel=10))
            with assert_raises(ValueError):
                cls.normalize_opts(dict(shuffle=3))


class TestZlibCompressor(unittest.TestCase, CompressorTests):

    compression = 'zlib'

    def test_normalize_opts(self):
        cls = get_compressor_cls(self.compression)

        # test defaults
        opts = cls.normalize_opts(None)
        eq(cls.default_level, opts)

        # test invalid args
        with assert_raises(ValueError):
            cls.normalize_opts(10)


class TestBZ2Compressor(unittest.TestCase, CompressorTests):

    compression = 'bz2'

    def test_normalize_opts(self):
        cls = get_compressor_cls(self.compression)

        # test defaults
        opts = cls.normalize_opts(None)
        eq(cls.default_level, opts)

        # test invalid args
        with assert_raises(ValueError):
            cls.normalize_opts(10)


try:
    import lzma
except ImportError:
    print('LZMA not available, skipping LZMA compressor tests')
else:

    class TestLZMACompressor(unittest.TestCase, CompressorTests):

        compression = 'lzma'

        def test_normalize_opts(self):
            cls = get_compressor_cls(self.compression)

            # test defaults
            opts = cls.normalize_opts(None)
            default_opts = dict(
                format=cls.default_format,
                check=cls.default_check,
                preset=cls.default_preset,
                filters=cls.default_filters,
            )
            eq(default_opts, opts)

            # test invalid args
            with assert_raises(ValueError):
                cls.normalize_opts(dict(format=100))
            with assert_raises(ValueError):
                cls.normalize_opts(dict(check=100))
            with assert_raises(ValueError):
                cls.normalize_opts(dict(preset=100))

            # test preset overridden if filters
            opts_with_filter = cls.normalize_opts(
                dict(preset=1, filters=[dict(id=lzma.FILTER_LZMA2, preset=9)])
            )
            opts = cls.normalize_opts(opts_with_filter)
            eq(None, opts['preset'])

        def test_compress_decompress_raw(self):
            opts = dict(format=lzma.FORMAT_RAW,
                        filters=[dict(id=lzma.FILTER_LZMA2, preset=1)])
            self._test_compress_decompress(opts)


class TestNoCompressor(unittest.TestCase, CompressorTests):

    compression = 'none'

    def test_normalize_opts(self):
        cls = get_compressor_cls(self.compression)

        # test defaults
        opts = cls.normalize_opts(None)
        eq(None, opts)

        # test invalid args
        with assert_raises(ValueError):
            cls.normalize_opts(9)


def test_get_compressor_cls():
    # expect ValueError, more friendly
    with assert_raises(ValueError):
        get_compressor_cls('foo')



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, division
import json


from nose.tools import eq_ as eq, assert_is_none, assert_raises
import numpy as np


from zarr.meta import decode_metadata, encode_dtype, decode_dtype
from zarr.errors import MetadataError


def test_decode():

    # typical
    b = b'''{
        "zarr_format": 1,
        "shape": [100],
        "chunks": [10],
        "dtype": "<f8",
        "compression": "zlib",
        "compression_opts": 1,
        "fill_value": null,
        "order": "C"
    }'''
    meta = decode_metadata(b)
    eq(1, meta['zarr_format'])
    eq((100,), meta['shape'])
    eq((10,), meta['chunks'])
    eq(np.dtype('<f8'), meta['dtype'])
    eq('zlib', meta['compression'])
    eq(1, meta['compression_opts'])
    assert_is_none(meta['fill_value'])
    eq('C', meta['order'])

    # variations
    b = b'''{
        "zarr_format": 1,
        "shape": [100, 100],
        "chunks": [10, 10],
        "dtype": [["a", "i4"], ["b", "S10"]],
        "compression": "blosc",
        "compression_opts": {
            "cname": "lz4",
            "clevel": 3,
            "shuffle": 2
        },
        "fill_value": 42,
        "order": "F"
    }'''
    meta = decode_metadata(b)
    eq(1, meta['zarr_format'])
    eq((100, 100), meta['shape'])
    eq((10, 10), meta['chunks'])
    # check structured dtype
    eq(np.dtype([('a', 'i4'), ('b', 'S10')]), meta['dtype'])
    # check structured compression_opts
    eq(dict(cname='lz4', clevel=3, shuffle=2), meta['compression_opts'])
    # check fill value
    eq(42, meta['fill_value'])
    eq('F', meta['order'])

    # unsupported format
    b = b'''{
        "zarr_format": 2
    }'''
    with assert_raises(MetadataError):
        decode_metadata(b)

    # missing fields
    b = b'''{
        "zarr_format": 1
    }'''
    with assert_raises(MetadataError):
        decode_metadata(b)


def test_encode_decode_dtype():

    for dt in ['f8', [('a', 'f8')], [('a', 'f8'), ('b', 'i1')]]:
        e = encode_dtype(np.dtype(dt))
        s = json.dumps(e)  # check JSON serializable
        o = json.loads(s)
        d = decode_dtype(o)
        eq(np.dtype(dt), d)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, division
import unittest
import tempfile
import atexit
import shutil
import pickle
import json
import array


import numpy as np
from nose.tools import assert_raises, eq_ as eq, assert_is_none


from zarr.storage import DirectoryStore, init_store
from zarr.meta import decode_metadata
from zarr.compat import text_type


def test_init_store():

    store = dict()
    init_store(store, shape=1000, chunks=100)

    # check metadata
    assert 'meta' in store
    meta = decode_metadata(store['meta'])
    eq((1000,), meta['shape'])
    eq((100,), meta['chunks'])
    eq(np.dtype(None), meta['dtype'])
    eq('blosc', meta['compression'])
    assert 'compression_opts' in meta
    assert_is_none(meta['fill_value'])

    # check attributes
    assert 'attrs' in store
    eq(dict(), json.loads(text_type(store['attrs'], 'ascii')))


def test_init_store_overwrite():

    store = dict(shape=(2000,), chunks=(200,))

    # overwrite
    init_store(store, shape=1000, chunks=100, overwrite=True)
    assert 'meta' in store
    meta = decode_metadata(store['meta'])
    eq((1000,), meta['shape'])
    eq((100,), meta['chunks'])

    # don't overwrite
    with assert_raises(ValueError):
        init_store(store, shape=1000, chunks=100, overwrite=False)


class MappingTests(object):

    def create_mapping(self, **kwargs):
        pass

    def test_get_set_del_contains(self):
        m = self.create_mapping()
        assert 'foo' not in m
        m['foo'] = b'bar'
        assert 'foo' in m
        eq(b'bar', m['foo'])
        del m['foo']
        assert 'foo' not in m
        with assert_raises(KeyError):
            m['foo']
        with assert_raises(KeyError):
            del m['foo']
        with assert_raises(TypeError):
            # non-writeable value
            m['foo'] = 42
        # alternative values
        m['foo'] = bytearray(b'bar')
        eq(b'bar', m['foo'])
        m['foo'] = array.array('B', b'bar')
        eq(b'bar', m['foo'])

    def test_update(self):
        m = self.create_mapping()
        assert 'foo' not in m
        assert 'baz' not in m
        m.update(foo=b'bar', baz=b'quux')
        eq(b'bar', m['foo'])
        eq(b'quux', m['baz'])

    def test_iterators(self):
        m = self.create_mapping()
        eq(0, len(m))
        eq(set(), set(m))
        eq(set(), set(m.keys()))
        eq(set(), set(m.values()))
        eq(set(), set(m.items()))

        m['foo'] = b'bar'
        m['baz'] = b'quux'

        eq(2, len(m))
        eq(set(['foo', 'baz']), set(m))
        eq(set(['foo', 'baz']), set(m.keys()))
        eq(set([b'bar', b'quux']), set(m.values()))
        eq(set([('foo', b'bar'), ('baz', b'quux')]), set(m.items()))


class TestDirectoryMap(MappingTests, unittest.TestCase):

    def create_mapping(self, **kwargs):
        path = tempfile.mkdtemp()
        atexit.register(shutil.rmtree, path)
        m = DirectoryStore(path, **kwargs)
        return m

    def test_size(self):
        m = self.create_mapping()
        eq(0, m.size)
        m['foo'] = b'bar'
        eq(3, m.size)
        m['baz'] = b'quux'
        eq(7, m.size)

    def test_path(self):
        with assert_raises(ValueError):
            DirectoryStore('doesnotexist')
        with tempfile.NamedTemporaryFile() as f:
            with assert_raises(ValueError):
                DirectoryStore(f.name)

    def test_pickle(self):
        m = self.create_mapping()
        m['foo'] = b'bar'
        m['baz'] = b'quux'
        m2 = pickle.loads(pickle.dumps(m))
        eq(len(m), len(m2))
        eq(m.path, m2.path)
        eq(b'bar', m2['foo'])
        eq(b'quux', m2['baz'])
        assert 'xxx' not in m
        m2['xxx'] = b'yyy'
        eq(b'yyy', m['xxx'])



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, division
import os


import numpy as np


from zarr.core import Array
from zarr.sync import SynchronizedArray
from zarr.storage import DirectoryStore, init_store


def create(shape, chunks, dtype=None, compression='default',
           compression_opts=None, fill_value=None, order='C', store=None,
           synchronizer=None, overwrite=False):
    """Create an array.

    Parameters
    ----------
    shape : int or tuple of ints
        Array shape.
    chunks : int or tuple of ints
        Chunk shape.
    dtype : string or dtype, optional
        NumPy dtype.
    compression : string, optional
        Name of primary compression library, e.g., 'blosc', 'zlib', 'bz2',
        'lzma'.
    compression_opts : object, optional
        Options to primary compressor. E.g., for blosc, provide a dictionary
        with keys 'cname', 'clevel' and 'shuffle'.
    fill_value : object
        Default value to use for uninitialised portions of the array.
    order : {'C', 'F'}, optional
        Memory layout to be used within each chunk.
    store : MutableMapping, optional
        Array storage. If not provided, a Python dict will be used, meaning
        array data will be stored in memory.
    synchronizer : zarr.sync.ArraySynchronizer, optional
        Array synchronizer.
    overwrite : bool, optional
        If True, delete all pre-existing data in `store` before creating the
        array.

    Returns
    -------
    z : zarr.core.Array

    Examples
    --------

    Create an array with default settings::

        >>> import zarr
        >>> z = zarr.create((10000, 10000), chunks=(1000, 1000))
        >>> z
        zarr.core.Array((10000, 10000), float64, chunks=(1000, 1000), order=C)
          compression: blosc; compression_opts: {'clevel': 5, 'cname': 'blosclz', 'shuffle': 1}
          nbytes: 762.9M; nbytes_stored: 320; ratio: 2500000.0; initialized: 0/100
          store: builtins.dict

    """  # flake8: noqa

    # initialise store
    if store is None:
        store = dict()
    init_store(store, shape=shape, chunks=chunks, dtype=dtype,
               compression=compression, compression_opts=compression_opts,
               fill_value=fill_value, order=order, overwrite=overwrite)

    # instantiate array
    if synchronizer is not None:
        z = SynchronizedArray(store, synchronizer)
    else:
        z = Array(store)

    return z


def empty(shape, chunks, dtype=None, compression='default',
          compression_opts=None, order='C', store=None, synchronizer=None):
    """Create an empty array.

    For parameter definitions see :func:`zarr.creation.create`.

    Notes
    -----
    The contents of an empty Zarr array are not defined. On attempting to
    retrieve data from an empty Zarr array, any values may be returned,
    and these are not guaranteed to be stable from one access to the next.

    """
    return create(shape=shape, chunks=chunks, dtype=dtype,
                  compression=compression, compression_opts=compression_opts,
                  fill_value=None, order=order, store=store,
                  synchronizer=synchronizer)


def zeros(shape, chunks, dtype=None, compression='default',
          compression_opts=None, order='C', store=None, synchronizer=None):
    """Create an array, with zero being used as the default value for
    uninitialised portions of the array.

    For parameter definitions see :func:`zarr.creation.create`.

    Examples
    --------
    >>> import zarr
    >>> z = zarr.zeros((10000, 10000), chunks=(1000, 1000))
    >>> z
    zarr.core.Array((10000, 10000), float64, chunks=(1000, 1000), order=C)
      compression: blosc; compression_opts: {'clevel': 5, 'cname': 'blosclz', 'shuffle': 1}
      nbytes: 762.9M; nbytes_stored: 317; ratio: 2523659.3; initialized: 0/100
      store: builtins.dict
    >>> z[:2, :2]
    array([[ 0.,  0.],
           [ 0.,  0.]])

    """  # flake8: noqa

    return create(shape=shape, chunks=chunks, dtype=dtype,
                  compression=compression,
                  compression_opts=compression_opts, fill_value=0, order=order,
                  store=store, synchronizer=synchronizer)


def ones(shape, chunks, dtype=None, compression='default',
         compression_opts=None, order='C', store=None, synchronizer=None):
    """Create an array, with one being used as the default value for
    uninitialised portions of the array.

    For parameter definitions see :func:`zarr.creation.create`.

    Examples
    --------
    >>> import zarr
    >>> z = zarr.ones((10000, 10000), chunks=(1000, 1000))
    >>> z
    zarr.core.Array((10000, 10000), float64, chunks=(1000, 1000), order=C)
      compression: blosc; compression_opts: {'clevel': 5, 'cname': 'blosclz', 'shuffle': 1}
      nbytes: 762.9M; nbytes_stored: 317; ratio: 2523659.3; initialized: 0/100
      store: builtins.dict
    >>> z[:2, :2]
    array([[ 1.,  1.],
           [ 1.,  1.]])

    """  # flake8: noqa

    return create(shape=shape, chunks=chunks, dtype=dtype,
                  compression=compression, compression_opts=compression_opts,
                  fill_value=1, order=order, store=store,
                  synchronizer=synchronizer)


def full(shape, chunks, fill_value, dtype=None, compression='default',
         compression_opts=None, order='C', store=None, synchronizer=None):
    """Create an array, with `fill_value` being used as the default value for
    uninitialised portions of the array.

    For parameter definitions see :func:`zarr.creation.create`.

    Examples
    --------
    >>> import zarr
    >>> z = zarr.full((10000, 10000), chunks=(1000, 1000), fill_value=42)
    >>> z
    zarr.core.Array((10000, 10000), float64, chunks=(1000, 1000), order=C)
      compression: blosc; compression_opts: {'clevel': 5, 'cname': 'blosclz', 'shuffle': 1}
      nbytes: 762.9M; nbytes_stored: 318; ratio: 2515723.3; initialized: 0/100
      store: builtins.dict
    >>> z[:2, :2]
    array([[ 42.,  42.],
           [ 42.,  42.]])

    """  # flake8: noqa

    return create(shape=shape, chunks=chunks, dtype=dtype,
                  compression=compression, compression_opts=compression_opts,
                  fill_value=fill_value, order=order, store=store,
                  synchronizer=synchronizer)


def array(data, chunks=None, dtype=None, compression='default',
          compression_opts=None, fill_value=None, order='C', store=None,
          synchronizer=None):
    """Create an array filled with `data`.

    The `data` argument should be a NumPy array or array-like object. For
    other parameter definitions see :func:`zarr.creation.create`.

    Examples
    --------
    >>> import numpy as np
    >>> import zarr
    >>> a = np.arange(100000000).reshape(10000, 10000)
    >>> z = zarr.array(a, chunks=(1000, 1000))
    >>> z
    zarr.core.Array((10000, 10000), int64, chunks=(1000, 1000), order=C)
      compression: blosc; compression_opts: {'clevel': 5, 'cname': 'blosclz', 'shuffle': 1}
      nbytes: 762.9M; nbytes_stored: 16.8M; ratio: 45.5; initialized: 100/100
      store: builtins.dict

    """  # flake8: noqa

    # ensure data is array-like
    if not hasattr(data, 'shape') or not hasattr(data, 'dtype'):
        data = np.asanyarray(data)

    # setup dtype
    if dtype is None:
        dtype = data.dtype

    # setup shape
    shape = data.shape

    # setup chunks
    if chunks is None:
        # try to use same chunks as data
        if hasattr(data, 'chunklen'):
            # bcolz carray
            chunks = (data.chunklen,) + shape[1:]
        elif hasattr(data, 'chunks') and len(data.chunks) == len(data.shape):
            # h5py dataset or zarr array
            chunks = data.chunks
        else:
            raise ValueError('chunks must be specified')

    # instantiate array
    z = create(shape=shape, chunks=chunks, dtype=dtype,
               compression=compression, compression_opts=compression_opts,
               fill_value=fill_value, order=order, store=store,
               synchronizer=synchronizer)

    # fill with data
    z[:] = data

    return z


# noinspection PyShadowingBuiltins
def open(path, mode='a', shape=None, chunks=None, dtype=None,
         compression='default', compression_opts=None, fill_value=0, order='C',
         synchronizer=None):
    """Open an array stored in a directory on the file system.

    Parameters
    ----------
    path : string
        Path to directory in which to store the array.
    mode : {'r', 'r+', 'a', 'w', 'w-'}
        Persistence mode: 'r' means readonly (must exist); 'r+' means
        read/write (must exist); 'a' means read/write (create if doesn't
        exist); 'w' means create (overwrite if exists); 'w-' means create
        (fail if exists).
    shape : int or tuple of ints
        Array shape.
    chunks : int or tuple of ints
        Chunk shape.
    dtype : string or dtype, optional
        NumPy dtype.
    compression : string, optional
        Name of primary compression library, e.g., 'blosc', 'zlib', 'bz2',
        'lzma'.
    compression_opts : object, optional
        Options to primary compressor. E.g., for blosc, provide a dictionary
        with keys 'cname', 'clevel' and 'shuffle'.
    fill_value : object
        Default value to use for uninitialised portions of the array.
    order : {'C', 'F'}, optional
        Memory layout to be used within each chunk.
    synchronizer : zarr.sync.ArraySynchronizer, optional
        Array synchronizer.

    Returns
    -------
    z : zarr.core.Array

    Examples
    --------
    >>> import numpy as np
    >>> import zarr
    >>> z1 = zarr.open('example.zarr', mode='w', shape=(10000, 10000),
    ...                chunks=(1000, 1000), fill_value=0)
    >>> z1[:] = np.arange(100000000).reshape(10000, 10000)
    >>> z1
    zarr.core.Array((10000, 10000), float64, chunks=(1000, 1000), order=C)
      compression: blosc; compression_opts: {'clevel': 5, 'cname': 'blosclz', 'shuffle': 1}
      nbytes: 762.9M; nbytes_stored: 24.6M; ratio: 31.0; initialized: 100/100
      store: zarr.storage.DirectoryStore
    >>> z2 = zarr.open('example.zarr', mode='r')
    >>> z2
    zarr.core.Array((10000, 10000), float64, chunks=(1000, 1000), order=C)
      compression: blosc; compression_opts: {'clevel': 5, 'cname': 'blosclz', 'shuffle': 1}
      nbytes: 762.9M; nbytes_stored: 24.6M; ratio: 31.0; initialized: 100/100
      store: zarr.storage.DirectoryStore
    >>> np.all(z1[:] == z2[:])
    True

    Notes
    -----
    There is no need to close an array. Data are automatically flushed to the
    file system.

    """  # flake8: noqa

    # use same mode semantics as h5py, although N.B., here `path` is a
    # directory:
    # r : readonly, must exist
    # r+ : read/write, must exist
    # w : create, delete if exists
    # w- or x : create, fail if exists
    # a : read/write if exists, create otherwise (default)

    # ensure directory exists
    if not os.path.exists(path):
        if mode in ['w', 'w-', 'x', 'a']:
            os.makedirs(path)
        elif mode in ['r', 'r+']:
            raise ValueError('path does not exist: %r' % path)

    # setup store
    store = DirectoryStore(path)
    exists = 'meta' in store  # use metadata key as indicator of existence

    # ensure store is initialized
    if mode in ['r', 'r+'] and not exists:
        raise ValueError('array does not exist')
    elif mode in ['w-', 'x'] and exists:
        raise ValueError('array exists')
    elif mode == 'w' or (mode in ['a', 'w-', 'x'] and not exists):
        init_store(store, shape=shape, chunks=chunks, dtype=dtype,
                   compression=compression, compression_opts=compression_opts,
                   fill_value=fill_value, order=order, overwrite=True)

    # determine readonly status
    readonly = mode == 'r'

    # handle optional synchronizer
    if synchronizer is not None:
        z = SynchronizedArray(store, synchronizer, readonly=readonly)
    else:
        z = Array(store, readonly=readonly)

    return z


def _like_args(a, shape, chunks, dtype, compression, compression_opts, order):
    if shape is None:
        shape = a.shape
    if chunks is None:
        try:
            chunks = a.chunks
        except AttributeError:
            raise ValueError('chunks must be specified')
    if dtype is None:
        dtype = a.dtype
    if compression is None:
        try:
            compression = a.compression
        except AttributeError:
            compression = 'blosc'
    if compression_opts is None:
        try:
            compression_opts = a.compression_opts
        except AttributeError:
            pass
    if order is None:
        try:
            order = a.order
        except AttributeError:
            order = 'C'
    return shape, chunks, dtype, compression, compression_opts, order


def empty_like(a, shape=None, chunks=None, dtype=None, compression=None,
               compression_opts=None, order=None, store=None,
               synchronizer=None):
    """Create an empty array like `a`."""
    shape, chunks, dtype, compression, compression_opts, order = \
        _like_args(a, shape, chunks, dtype, compression, compression_opts,
                   order)
    return empty(shape, chunks, dtype=dtype, compression=compression,
                 compression_opts=compression_opts, order=order,
                 store=store, synchronizer=synchronizer)


def zeros_like(a, shape=None, chunks=None, dtype=None, compression=None,
               compression_opts=None, order=None, store=None,
               synchronizer=None):
    """Create an array of zeros like `a`."""
    shape, chunks, dtype, compression, compression_opts, order = \
        _like_args(a, shape, chunks, dtype, compression, compression_opts,
                   order)
    return zeros(shape, chunks, dtype=dtype, compression=compression,
                 compression_opts=compression_opts, order=order,
                 store=store, synchronizer=synchronizer)


def ones_like(a, shape=None, chunks=None, dtype=None, compression=None,
              compression_opts=None, order=None, store=None,
              synchronizer=None):
    """Create an array of ones like `a`."""
    shape, chunks, dtype, compression, compression_opts, order = \
        _like_args(a, shape, chunks, dtype, compression, compression_opts,
                   order)
    return ones(shape, chunks, dtype=dtype, compression=compression,
                compression_opts=compression_opts, order=order,
                store=store, synchronizer=synchronizer)


def full_like(a, shape=None, chunks=None, fill_value=None, dtype=None,
              compression=None, compression_opts=None, order=None,
              store=None, synchronizer=None):
    """Create a filled array like `a`."""
    shape, chunks, dtype, compression, compression_opts, order = \
        _like_args(a, shape, chunks, dtype, compression, compression_opts,
                   order)
    if fill_value is None:
        try:
            fill_value = a.fill_value
        except AttributeError:
            raise ValueError('fill_value must be specified')
    return full(shape, chunks, fill_value, dtype=dtype,
                compression=compression, compression_opts=compression_opts,
                order=order, store=store, synchronizer=synchronizer)


def open_like(a, path, mode='a', shape=None, chunks=None, dtype=None,
              compression=None, compression_opts=None, fill_value=None,
              order=None, synchronizer=None):
    """Open a persistent array like `a`."""
    shape, chunks, dtype, compression, compression_opts, order = \
        _like_args(a, shape, chunks, dtype, compression, compression_opts,
                   order)
    if fill_value is None:
        try:
            fill_value = a.fill_value
        except AttributeError:
            # leave empty
            pass
    return open(path, mode=mode, shape=shape, chunks=chunks, dtype=dtype,
                compression=compression, compression_opts=compression_opts,
                fill_value=fill_value, order=order, synchronizer=synchronizer)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, division
import tempfile
import shutil
import atexit
import os


import numpy as np
from nose.tools import eq_ as eq, assert_is_none, assert_is_instance, \
    assert_raises
from numpy.testing import assert_array_equal


from zarr.creation import array, empty, zeros, ones, full, open, empty_like, \
    zeros_like, ones_like, full_like, open_like, create
from zarr.sync import ThreadSynchronizer, SynchronizedArray
from zarr.core import Array
from zarr.storage import DirectoryStore, init_store


def test_array():

    # with numpy array
    a = np.arange(100)
    z = array(a, chunks=10)
    eq(a.shape, z.shape)
    eq(a.dtype, z.dtype)
    assert_array_equal(a, z[:])

    # with array-like
    a = list(range(100))
    z = array(a, chunks=10)
    eq((100,), z.shape)
    eq(np.asarray(a).dtype, z.dtype)
    assert_array_equal(np.asarray(a), z[:])

    # with another zarr array
    z2 = array(z)
    eq(z.shape, z2.shape)
    eq(z.chunks, z2.chunks)
    eq(z.dtype, z2.dtype)
    assert_array_equal(z[:], z2[:])

    # with something bcolz-like
    class MockBcolzArray(object):

        def __init__(self, data, chunklen):
            self.data = data
            self.chunklen = chunklen

        def __getattr__(self, item):
            return getattr(self.data, item)

        def __getitem__(self, item):
            return self.data[item]

    b = np.arange(1000).reshape(100, 10)
    c = MockBcolzArray(b, 10)
    z3 = array(c)
    eq(c.shape, z3.shape)
    eq((10, 10), z3.chunks)

    # chunks not specified
    with assert_raises(ValueError):
        z = array(np.arange(100))


def test_empty():
    z = empty(100, 10)
    eq((100,), z.shape)
    eq((10,), z.chunks)


def test_zeros():
    z = zeros(100, 10)
    eq((100,), z.shape)
    eq((10,), z.chunks)
    assert_array_equal(np.zeros(100), z[:])


def test_ones():
    z = ones(100, 10)
    eq((100,), z.shape)
    eq((10,), z.chunks)
    assert_array_equal(np.ones(100), z[:])


def test_full():
    z = full(100, 10, fill_value=42, dtype='i4')
    eq((100,), z.shape)
    eq((10,), z.chunks)
    assert_array_equal(np.full(100, fill_value=42, dtype='i4'), z[:])


def test_open():

    path = tempfile.mktemp()
    atexit.register(
        lambda: shutil.rmtree(path) if os.path.exists(path) else None
    )
    z = open(path, mode='w', shape=100, chunks=10, dtype='i4')
    z[:] = 42
    eq((100,), z.shape)
    eq((10,), z.chunks)
    assert_array_equal(np.full(100, fill_value=42, dtype='i4'), z[:])
    z2 = open(path, mode='r')
    eq((100,), z2.shape)
    eq((10,), z2.chunks)
    assert_array_equal(z[:], z2[:])

    # path does not exist
    path = 'doesnotexist'
    with assert_raises(ValueError):
        open(path, mode='r')

    # path exists but store not initialised
    path = tempfile.mkdtemp()
    atexit.register(shutil.rmtree, path)
    with assert_raises(ValueError):
        open(path, mode='r')
    with assert_raises(ValueError):
        open(path, mode='r+')

    # store initialised, mode w-
    store = DirectoryStore(path)
    init_store(store, shape=100, chunks=10)
    with assert_raises(ValueError):
        open(path, mode='w-')
    with assert_raises(ValueError):
        open(path, mode='x')

    # with synchronizer
    z = open(path, synchronizer=ThreadSynchronizer())
    assert_is_instance(z, SynchronizedArray)


def test_empty_like():
    # zarr array
    z = empty(100, 10, dtype='f4', compression='zlib',
              compression_opts=5, order='F')
    z2 = empty_like(z)
    eq(z.shape, z2.shape)
    eq(z.chunks, z2.chunks)
    eq(z.dtype, z2.dtype)
    eq(z.compression, z2.compression)
    eq(z.compression_opts, z2.compression_opts)
    eq(z.fill_value, z2.fill_value)
    eq(z.order, z2.order)
    # numpy array
    a = np.empty(100, dtype='f4')
    z3 = empty_like(a, chunks=10)
    eq(a.shape, z3.shape)
    eq((10,), z3.chunks)
    eq(a.dtype, z3.dtype)
    assert_is_none(z3.fill_value)
    with assert_raises(ValueError):
        # chunks missing
        empty_like(a)


def test_zeros_like():
    # zarr array
    z = zeros(100, 10, dtype='f4', compression='zlib',
              compression_opts=5, order='F')
    z2 = zeros_like(z)
    eq(z.shape, z2.shape)
    eq(z.chunks, z2.chunks)
    eq(z.dtype, z2.dtype)
    eq(z.compression, z2.compression)
    eq(z.compression_opts, z2.compression_opts)
    eq(z.fill_value, z2.fill_value)
    eq(z.order, z2.order)
    # numpy array
    a = np.empty(100, dtype='f4')
    z3 = zeros_like(a, chunks=10)
    eq(a.shape, z3.shape)
    eq((10,), z3.chunks)
    eq(a.dtype, z3.dtype)
    eq(0, z3.fill_value)
    with assert_raises(ValueError):
        # chunks missing
        zeros_like(a)


def test_ones_like():
    # zarr array
    z = ones(100, 10, dtype='f4', compression='zlib',
             compression_opts=5, order='F')
    z2 = ones_like(z)
    eq(z.shape, z2.shape)
    eq(z.chunks, z2.chunks)
    eq(z.dtype, z2.dtype)
    eq(z.compression, z2.compression)
    eq(z.compression_opts, z2.compression_opts)
    eq(z.fill_value, z2.fill_value)
    eq(z.order, z2.order)
    # numpy array
    a = np.empty(100, dtype='f4')
    z3 = ones_like(a, chunks=10)
    eq(a.shape, z3.shape)
    eq((10,), z3.chunks)
    eq(a.dtype, z3.dtype)
    eq(1, z3.fill_value)
    with assert_raises(ValueError):
        # chunks missing
        ones_like(a)


def test_full_like():
    z = full(100, 10, dtype='f4', compression='zlib',
             compression_opts=5, fill_value=42, order='F')
    z2 = full_like(z)
    eq(z.shape, z2.shape)
    eq(z.chunks, z2.chunks)
    eq(z.dtype, z2.dtype)
    eq(z.compression, z2.compression)
    eq(z.compression_opts, z2.compression_opts)
    eq(z.fill_value, z2.fill_value)
    eq(z.order, z2.order)
    # numpy array
    a = np.empty(100, dtype='f4')
    z3 = full_like(a, chunks=10, fill_value=42)
    eq(a.shape, z3.shape)
    eq((10,), z3.chunks)
    eq(a.dtype, z3.dtype)
    eq(42, z3.fill_value)
    with assert_raises(ValueError):
        # chunks missing
        full_like(a)
    with assert_raises(ValueError):
        # fill_value missing
        full_like(a, chunks=10)


def test_open_like():
    # zarr array
    path = tempfile.mktemp()
    atexit.register(shutil.rmtree, path)
    z = full(100, 10, dtype='f4', compression='zlib',
             compression_opts=5, fill_value=42, order='F')
    z2 = open_like(z, path)
    eq(z.shape, z2.shape)
    eq(z.chunks, z2.chunks)
    eq(z.dtype, z2.dtype)
    eq(z.compression, z2.compression)
    eq(z.compression_opts, z2.compression_opts)
    eq(z.fill_value, z2.fill_value)
    eq(z.order, z2.order)
    # numpy array
    path = tempfile.mktemp()
    atexit.register(shutil.rmtree, path)
    a = np.empty(100, dtype='f4')
    z3 = open_like(a, path, chunks=10)
    eq(a.shape, z3.shape)
    eq((10,), z3.chunks)
    eq(a.dtype, z3.dtype)
    assert_is_none(z3.fill_value)
    with assert_raises(ValueError):
        # chunks missing
        open_like(a, path)


def test_create():

    # defaults
    z = create(100, 10)
    assert_is_instance(z, Array)
    eq((100,), z.shape)
    eq((10,), z.chunks)
    eq(np.dtype(None), z.dtype)
    eq('blosc', z.compression)
    assert_is_none(z.fill_value)

    # all specified
    z = create(100, 10, dtype='i4', compression='zlib', compression_opts=1,
               fill_value=42, order='F')
    assert_is_instance(z, Array)
    eq((100,), z.shape)
    eq((10,), z.chunks)
    eq(np.dtype('i4'), z.dtype)
    eq('zlib', z.compression)
    eq(1, z.compression_opts)
    eq(42, z.fill_value)
    eq('F', z.order)

    # with synchronizer
    synchronizer = ThreadSynchronizer()
    z = create(100, 10, synchronizer=synchronizer)
    assert_is_instance(z, SynchronizedArray)
    eq((100,), z.shape)
    eq((10,), z.chunks)
    assert synchronizer is z.synchronizer



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, division
import json
import unittest


from nose.tools import eq_ as eq, assert_raises


from zarr.attrs import Attributes
from zarr.compat import binary_type, text_type
from zarr.errors import ReadOnlyError


class TestAttributes(unittest.TestCase):

    def init_attributes(self, store, readonly=False):
        return Attributes(store, readonly=readonly)

    def test_storage(self):

        store = dict()
        a = self.init_attributes(store)
        assert 'attrs' in store
        assert isinstance(store['attrs'], binary_type)
        d = json.loads(text_type(store['attrs'], 'ascii'))
        eq(dict(), d)

        a['foo'] = 'bar'
        a['baz'] = 42

        d = json.loads(text_type(store['attrs'], 'ascii'))
        eq(dict(foo='bar', baz=42), d)

    def test_get_set_del_contains(self):

        a = self.init_attributes(dict())
        assert 'foo' not in a
        a['foo'] = 'bar'
        a['baz'] = 42
        assert 'foo' in a
        assert 'baz' in a
        eq('bar', a['foo'])
        eq(42, a['baz'])
        del a['foo']
        assert 'foo' not in a
        with assert_raises(KeyError):
            a['foo']

    def test_update(self):

        a = self.init_attributes(dict())
        assert 'foo' not in a
        assert 'baz' not in a
        a.update(foo='bar', baz=42)
        eq(a['foo'], 'bar')
        eq(a['baz'], 42)

    def test_iterators(self):

        a = self.init_attributes(dict())
        eq(0, len(a))
        eq(set(), set(a))
        eq(set(), set(a.keys()))
        eq(set(), set(a.values()))
        eq(set(), set(a.items()))

        a['foo'] = 'bar'
        a['baz'] = 42

        eq(2, len(a))
        eq({'foo', 'baz'}, set(a))
        eq({'foo', 'baz'}, set(a.keys()))
        eq({'bar', 42}, set(a.values()))
        eq({('foo', 'bar'), ('baz', 42)}, set(a.items()))

    def test_readonly(self):

        store = dict()
        store['attrs'] = json.dumps(dict(foo='bar', baz=42)).encode('ascii')
        a = self.init_attributes(store, readonly=True)
        eq(a['foo'], 'bar')
        eq(a['baz'], 42)
        with assert_raises(ReadOnlyError):
            a['foo'] = 'quux'
        with assert_raises(ReadOnlyError):
            del a['foo']
        with assert_raises(ReadOnlyError):
            a.update(foo='quux')
        with assert_raises(ReadOnlyError):
            a.put(dict())



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, division


from nose.tools import eq_ as eq, assert_raises, assert_true, assert_false


from zarr.util import normalize_shape, normalize_chunks, is_total_slice, \
    normalize_axis_selection, normalize_array_selection, \
    normalize_resize_args, human_readable_size, normalize_order


def test_normalize_shape():
    eq((100,), normalize_shape((100,)))
    eq((100,), normalize_shape([100]))
    eq((100,), normalize_shape(100))
    with assert_raises(TypeError):
        normalize_shape(None)
    with assert_raises(ValueError):
        normalize_shape('foo')


def test_normalize_chunks():
    eq((10,), normalize_chunks((10,), (100,)))
    eq((10,), normalize_chunks([10], (100,)))
    eq((10,), normalize_chunks(10, (100,)))
    eq((10, 10), normalize_chunks((10, 10), (100, 10)))
    eq((10, 10), normalize_chunks(10, (100, 10)))
    eq((10, 10), normalize_chunks((10, None), (100, 10)))
    with assert_raises(TypeError):
        normalize_chunks(None, (100,))
    with assert_raises(ValueError):
        normalize_chunks('foo', (100,))
    with assert_raises(ValueError):
        normalize_chunks((100, 10), (100,))


def test_is_total_slice():

    # 1D
    assert_true(is_total_slice(Ellipsis, (100,)))
    assert_true(is_total_slice(slice(None), (100,)))
    assert_true(is_total_slice(slice(0, 100), (100,)))
    assert_false(is_total_slice(slice(0, 50), (100,)))
    assert_false(is_total_slice(slice(0, 100, 2), (100,)))

    # 2D
    assert_true(is_total_slice(Ellipsis, (100, 100)))
    assert_true(is_total_slice(slice(None), (100, 100)))
    assert_true(is_total_slice((slice(None), slice(None)), (100, 100)))
    assert_true(is_total_slice((slice(0, 100), slice(0, 100)), (100, 100)))
    assert_false(is_total_slice((slice(0, 100), slice(0, 50)), (100, 100)))
    assert_false(is_total_slice((slice(0, 50), slice(0, 100)), (100, 100)))
    assert_false(is_total_slice((slice(0, 50), slice(0, 50)), (100, 100)))
    assert_false(is_total_slice((slice(0, 100, 2), slice(0, 100)), (100, 100)))

    with assert_raises(TypeError):
        is_total_slice('foo', (100,))


def test_normalize_axis_selection():

    # single item
    eq(1, normalize_axis_selection(1, 100))
    eq(99, normalize_axis_selection(-1, 100))
    with assert_raises(IndexError):
        normalize_axis_selection(100, 100)
    with assert_raises(IndexError):
        normalize_axis_selection(1000, 100)
    with assert_raises(IndexError):
        normalize_axis_selection(-1000, 100)

    # slice
    eq(slice(0, 100), normalize_axis_selection(slice(None), 100))
    eq(slice(0, 100), normalize_axis_selection(slice(None, 100), 100))
    eq(slice(0, 100), normalize_axis_selection(slice(0, None), 100))
    eq(slice(0, 100), normalize_axis_selection(slice(0, 1000), 100))
    eq(slice(99, 100), normalize_axis_selection(slice(-1, None), 100))
    eq(slice(98, 99), normalize_axis_selection(slice(-2, -1), 100))
    with assert_raises(IndexError):
        normalize_axis_selection(slice(100, None), 100)
    with assert_raises(IndexError):
        normalize_axis_selection(slice(1000, 2000), 100)
    with assert_raises(IndexError):
        normalize_axis_selection(slice(-1000, 0), 100)
    with assert_raises(IndexError):
        normalize_axis_selection(slice(10, 0), 100)

    with assert_raises(TypeError):
        normalize_axis_selection('foo', 100)

    with assert_raises(NotImplementedError):
        normalize_axis_selection(slice(0, 100, 2), 100)


def test_normalize_array_selection():

    # 1D, single item
    eq((0,), normalize_array_selection(0, (100,)))

    # 1D, slice
    eq((slice(0, 100),), normalize_array_selection(Ellipsis, (100,)))
    eq((slice(0, 100),), normalize_array_selection(slice(None), (100,)))
    eq((slice(0, 100),), normalize_array_selection(slice(None, 100), (100,)))
    eq((slice(0, 100),), normalize_array_selection(slice(0, None), (100,)))

    # 2D, single item
    eq((0, 0), normalize_array_selection((0, 0), (100, 100)))
    eq((99, 1), normalize_array_selection((-1, 1), (100, 100)))

    # 2D, single col/row
    eq((0, slice(0, 100)), normalize_array_selection((0, slice(None)),
                                                     (100, 100)))
    eq((0, slice(0, 100)), normalize_array_selection((0,),
                                                     (100, 100)))
    eq((slice(0, 100), 0), normalize_array_selection((slice(None), 0),
                                                     (100, 100)))

    # 2D slice
    eq((slice(0, 100), slice(0, 100)),
       normalize_array_selection(Ellipsis, (100, 100)))
    eq((slice(0, 100), slice(0, 100)),
       normalize_array_selection(slice(None), (100, 100)))
    eq((slice(0, 100), slice(0, 100)),
       normalize_array_selection((slice(None), slice(None)), (100, 100)))

    with assert_raises(TypeError):
        normalize_array_selection('foo', (100,))


def test_normalize_resize_args():

    # 1D
    eq((200,), normalize_resize_args((100,), 200))
    eq((200,), normalize_resize_args((100,), (200,)))

    # 2D
    eq((200, 100), normalize_resize_args((100, 100), (200, 100)))
    eq((200, 100), normalize_resize_args((100, 100), (200, None)))
    eq((200, 100), normalize_resize_args((100, 100), 200, 100))
    eq((200, 100), normalize_resize_args((100, 100), 200, None))

    with assert_raises(ValueError):
        normalize_resize_args((100,), (200, 100))


def test_human_readable_size():
    eq('100', human_readable_size(100))
    eq('1.0K', human_readable_size(2**10))
    eq('1.0M', human_readable_size(2**20))
    eq('1.0G', human_readable_size(2**30))
    eq('1.0T', human_readable_size(2**40))
    eq('1.0P', human_readable_size(2**50))


def test_normalize_order():
    eq('F', normalize_order('F'))
    eq('C', normalize_order('C'))
    eq('F', normalize_order('f'))
    eq('C', normalize_order('c'))
    with assert_raises(ValueError):
        normalize_order('foo')



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, division
from threading import Lock
from collections import defaultdict
import os


import fasteners


from zarr.core import Array
from zarr.attrs import Attributes


class ThreadSynchronizer(object):
    """Provides synchronization using thread locks."""

    def __init__(self):
        self.mutex = Lock()
        self.attrs_lock = Lock()
        self.chunk_locks = defaultdict(Lock)

    def chunk_lock(self, ckey):
        with self.mutex:
            lock = self.chunk_locks[ckey]
        return lock

    def __getstate__(self):
        return dict()

    def __setstate__(self, d):
        # reinitialise from scratch
        self.mutex = Lock()
        self.attrs_lock = Lock()
        self.chunk_locks = defaultdict(Lock)


class ProcessSynchronizer(object):
    """Provides synchronization using file locks via the
    `fasteners <http://fasteners.readthedocs.io/en/latest/api/process_lock.html>`_
    package.

    Parameters
    ----------
    path : string
        Path to a directory on a file system that is shared by all processes.

    """  # flake8: noqa

    def __init__(self, path):
        self.path = path

    @property
    def attrs_lock(self):
        return fasteners.InterProcessLock(
            os.path.join(self.path, 'attrs.lock')
        )

    def chunk_lock(self, ckey):
        lock = fasteners.InterProcessLock(
            os.path.join(self.path, '%s.lock' % ckey)
        )
        return lock


class SynchronizedArray(Array):
    """Instantiate a synchronized array.

    Parameters
    ----------
    store : MutableMapping
        Array store, already initialised.
    synchronizer : object
        Array synchronizer.
    readonly : bool, optional
        True if array should be protected against modification.

    Examples
    --------
    >>> import zarr
    >>> store = dict()
    >>> zarr.init_store(store, shape=1000, chunks=100)
    >>> synchronizer = zarr.ThreadSynchronizer()
    >>> z = zarr.SynchronizedArray(store, synchronizer)
    >>> z
    zarr.sync.SynchronizedArray((1000,), float64, chunks=(100,), order=C)
      compression: blosc; compression_opts: {'clevel': 5, 'cname': 'blosclz', 'shuffle': 1}
      nbytes: 7.8K; nbytes_stored: 289; ratio: 27.7; initialized: 0/10
      store: builtins.dict; synchronizer: zarr.sync.ThreadSynchronizer

    Notes
    -----
    Only writing data to the array via the __setitem__() method and
    modification of user attributes are synchronized. Neither append() nor
    resize() are synchronized.

    Writing to the array is synchronized at the chunk level. I.e.,
    the array supports concurrent write operations via the __setitem__()
    method, but these will only exclude each other if they both require
    modification of the same chunk.

    """  # flake8: noqa

    def __init__(self, store, synchronizer, readonly=False):
        super(SynchronizedArray, self).__init__(store, readonly=readonly)
        self.synchronizer = synchronizer
        self._attrs = SynchronizedAttributes(store, synchronizer,
                                             readonly=readonly)

    def _chunk_setitem(self, cidx, key, value):
        ckey = '.'.join(map(str, cidx))
        with self.synchronizer.chunk_lock(ckey):
            super(SynchronizedArray, self)._chunk_setitem(cidx, key, value)

    def __repr__(self):
        r = super(SynchronizedArray, self).__repr__()
        r += ('; synchronizer: %s.%s' %
              (type(self.synchronizer).__module__,
               type(self.synchronizer).__name__))
        return r

    def __getstate__(self):
        return self._store, self.synchronizer, self._readonly

    def __setstate__(self, state):
        self.__init__(*state)


class SynchronizedAttributes(Attributes):

    def __init__(self, store, synchronizer, key='attrs', readonly=False):
        super(SynchronizedAttributes, self).__init__(store, key=key,
                                                     readonly=readonly)
        self.synchronizer = synchronizer

    def __setitem__(self, key, value):
        with self.synchronizer.attrs_lock:
            super(SynchronizedAttributes, self).__setitem__(key, value)

    def __delitem__(self, key):
        with self.synchronizer.attrs_lock:
            super(SynchronizedAttributes, self).__delitem__(key)

    def update(self, *args, **kwargs):
        with self.synchronizer.attrs_lock:
            super(SynchronizedAttributes, self).update(*args, **kwargs)



# -*- coding: utf-8 -*-
# flake8: noqa
from __future__ import absolute_import, print_function, division
import multiprocessing
import atexit


from zarr.creation import create, array, empty, zeros, ones, full, open, \
    empty_like, zeros_like, ones_like, full_like, open_like
from zarr.storage import init_store, DirectoryStore
from zarr.core import Array
from zarr.sync import ThreadSynchronizer, ProcessSynchronizer, \
    SynchronizedArray
from zarr.version import version as __version__


try:
    from zarr import blosc
except ImportError:  # pragma: no cover
    pass
else:
    ncores = multiprocessing.cpu_count()
    blosc.init()
    # diminishing returns beyond 4 threads?
    blosc.set_nthreads(min(4, ncores))
    atexit.register(blosc.destroy)



import numpy as np
import sys
sys.path.insert(0, '..')
import zarr
import line_profiler
import timeit
from zarr import blosc

# setup
a = np.random.normal(2000, 1000, size=200000000).astype('u2')
z = zarr.empty_like(a, chunks=1000000, compression='blosc', compression_opts=dict(cname='lz4', clevel=5, shuffle=2))
print(z)

print('*' * 79)

# time
t = timeit.repeat('z[:] = a', repeat=10, number=1, globals=globals())
print(t)
print(min(t))
print(z)

# profile
profile = line_profiler.LineProfiler(blosc.compress)
profile.run('z[:] = a')
profile.print_stats()

print('*' * 79)

# time
t = timeit.repeat('z[:]', repeat=10, number=1, globals=globals())
print(t)
print(min(t))

# profile
profile = line_profiler.LineProfiler(blosc.decompress)
profile.run('z[:]')
profile.print_stats()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, division
import json
from collections import MutableMapping


from zarr.compat import text_type
from zarr.errors import ReadOnlyError


class Attributes(MutableMapping):

    def __init__(self, store, key='attrs', readonly=False):
        if key not in store:
            store[key] = json.dumps(dict()).encode('ascii')
        self.store = store
        self.key = key
        self.readonly = readonly

    def __contains__(self, x):
        return x in self.asdict()

    def __getitem__(self, item):
        return self.asdict()[item]

    def put(self, d):

        # guard conditions
        if self.readonly:
            raise ReadOnlyError('attributes are read-only')

        s = json.dumps(d, indent=4, sort_keys=True, ensure_ascii=True)
        self.store[self.key] = s.encode('ascii')

    def __setitem__(self, key, value):

        # guard conditions
        if self.readonly:
            raise ReadOnlyError('attributes are read-only')

        # load existing data
        d = self.asdict()

        # set key value
        d[key] = value

        # put modified data
        self.put(d)

    def __delitem__(self, key):

        # guard conditions
        if self.readonly:
            raise ReadOnlyError('mapping is read-only')

        # load existing data
        d = self.asdict()

        # delete key value
        del d[key]

        # put modified data
        self.put(d)

    def asdict(self):
        return json.loads(text_type(self.store[self.key], 'ascii'))

    def update(self, *args, **kwargs):
        # override to provide update in a single write

        # guard conditions
        if self.readonly:
            raise ReadOnlyError('mapping is read-only')

        # load existing data
        d = self.asdict()

        # update
        d.update(*args, **kwargs)

        # put modified data
        self.put(d)

    def __iter__(self):
        return iter(self.asdict())

    def __len__(self):
        return len(self.asdict())

    def keys(self):
        return self.asdict().keys()

    def values(self):
        return self.asdict().values()

    def items(self):
        return self.asdict().items()



# -*- coding: utf-8 -*-
# cython: embedsignature=True
# cython: profile=False
# cython: linetrace=False
# cython: binding=False
from __future__ import absolute_import, print_function, division
import threading


# noinspection PyUnresolvedReferences
from cpython cimport array
import array
from cpython.buffer cimport PyObject_GetBuffer, PyBuffer_Release, \
    PyBUF_ANY_CONTIGUOUS, PyBUF_WRITEABLE


from zarr.compat import PY2, text_type


cdef extern from "blosc.h":
    cdef enum:
        BLOSC_MAX_OVERHEAD,
        BLOSC_VERSION_STRING,
        BLOSC_VERSION_DATE

    void blosc_init()
    void blosc_destroy()
    int blosc_set_nthreads(int nthreads)
    int blosc_set_compressor(const char *compname)
    int blosc_compress(int clevel, int doshuffle, size_t typesize,
                       size_t nbytes, void *src, void *dest,
                       size_t destsize) nogil
    int blosc_decompress(void *src, void *dest, size_t destsize) nogil
    int blosc_compname_to_compcode(const char *compname)
    int blosc_compress_ctx(int clevel, int doshuffle, size_t typesize,
                           size_t nbytes, const void* src, void* dest,
                           size_t destsize, const char* compressor,
				           size_t blocksize, int numinternalthreads) nogil
    int blosc_decompress_ctx(const void *src, void *dest, size_t destsize,
                             int numinternalthreads) nogil


def version():
    """Return the version of blosc that zarr was compiled with."""

    ver_str = <char *> BLOSC_VERSION_STRING
    ver_date = <char *> BLOSC_VERSION_DATE
    if not PY2:
        ver_str = ver_str.decode()
        ver_date = ver_date.decode()
    return ver_str, ver_date


def init():
    blosc_init()


def destroy():
    blosc_destroy()


def compname_to_compcode(cname):
    if isinstance(cname, text_type):
        cname = cname.encode('ascii')
    return blosc_compname_to_compcode(cname)


def set_nthreads(int nthreads):
    """Set the number of threads that Blosc uses internally for compression
    and decompression.
    ."""
    return blosc_set_nthreads(nthreads)


def decompress(source, dest):
    """Decompress data.

    Parameters
    ----------
    source : bytes-like
        Compressed data, including blosc header.
    dest : array-like
        Object to decompress into.

    Notes
    -----
    Assumes that the size of the destination buffer is correct for the size of
    the uncompressed data.

    """
    cdef:
        int ret
        char *source_ptr
        char *dest_ptr
        Py_buffer source_buffer
        array.array source_array
        Py_buffer dest_buffer
        size_t nbytes

    # setup source buffer
    if PY2 and isinstance(source, array.array):
        # workaround fact that array.array does not support new-style buffer
        # interface in PY2
        release_source_buffer = False
        source_array = source
        source_ptr = <char *> source_array.data.as_voidptr
    else:
        release_source_buffer = True
        PyObject_GetBuffer(source, &source_buffer, PyBUF_ANY_CONTIGUOUS)
        source_ptr = <char *> source_buffer.buf

    # setup destination buffer
    PyObject_GetBuffer(dest, &dest_buffer,
                       PyBUF_ANY_CONTIGUOUS | PyBUF_WRITEABLE)
    dest_ptr = <char *> dest_buffer.buf
    nbytes = dest_buffer.len

    # perform decompression
    if _get_use_threads():
        # allow blosc to use threads internally
        with nogil:
            ret = blosc_decompress(source_ptr, dest_ptr, nbytes)
    else:
        with nogil:
            ret = blosc_decompress_ctx(source_ptr, dest_ptr, nbytes, 1)

    # release buffers
    if release_source_buffer:
        PyBuffer_Release(&source_buffer)
    PyBuffer_Release(&dest_buffer)

    # handle errors
    if ret <= 0:
        raise RuntimeError('error during blosc decompression: %d' % ret)


def compress(source, char* cname, int clevel, int shuffle):
    """Compress data in a numpy array.

    Parameters
    ----------
    source : array-like
        Data to be compressed.
    cname : bytes
        Name of compression library to use.
    clevel : int
        Compression level.
    shuffle : int
        Shuffle filter.

    Returns
    -------
    dest : bytes-like
        Compressed data.

    """

    cdef:
        char *source_ptr
        char *dest_ptr
        Py_buffer source_buffer
        size_t nbytes, cbytes, itemsize
        array.array char_array_template = array.array('b', [])
        array.array dest

    # setup source buffer
    PyObject_GetBuffer(source, &source_buffer, PyBUF_ANY_CONTIGUOUS)
    source_ptr = <char *> source_buffer.buf

    # setup destination
    nbytes = source_buffer.len
    itemsize = source_buffer.itemsize
    dest = array.clone(char_array_template, nbytes + BLOSC_MAX_OVERHEAD,
                       zero=False)
    dest_ptr = <char *> dest.data.as_voidptr

    # perform compression
    if _get_use_threads():
        # allow blosc to use threads internally
        compressor_set = blosc_set_compressor(cname)
        if compressor_set < 0:
            raise ValueError('compressor not supported: %r' % cname)
        with nogil:
            cbytes = blosc_compress(clevel, shuffle, itemsize, nbytes,
                                    source_ptr, dest_ptr,
                                    nbytes + BLOSC_MAX_OVERHEAD)

    else:
        with nogil:
            cbytes = blosc_compress_ctx(clevel, shuffle, itemsize, nbytes,
                                        source_ptr, dest_ptr,
                                        nbytes + BLOSC_MAX_OVERHEAD, cname,
                                        0, 1)

    # release source buffer
    PyBuffer_Release(&source_buffer)

    # check compression was successful
    if cbytes <= 0:
        raise RuntimeError('error during blosc compression: %d' % cbytes)

    # resize after compression
    array.resize(dest, cbytes)

    return dest


# set the value of this variable to True or False to override the
# default adaptive behaviour
use_threads = None


def _get_use_threads():
    global use_threads

    if use_threads in [True, False]:
        # user has manually overridden the default behaviour
        _use_threads = use_threads

    else:
        # adaptive behaviour: allow blosc to use threads if it is being
        # called from the main Python thread, inferring that it is being run
        # from within a single-threaded program; otherwise do not allow
        # blosc to use threads, inferring it is being run from within a
        # multi-threaded program
        if hasattr(threading, 'main_thread'):
            _use_threads = (threading.main_thread() ==
                            threading.current_thread())
        else:
            _use_threads = threading.current_thread().name == 'MainThread'

    return _use_threads



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, division
from tempfile import mkdtemp
import atexit
import shutil


from zarr.tests.test_attrs import TestAttributes
from zarr.tests.test_core import TestArray
from zarr.sync import ThreadSynchronizer, ProcessSynchronizer, \
    SynchronizedArray, SynchronizedAttributes
from zarr.storage import init_store


class TestThreadSynchronizedAttributes(TestAttributes):

    def init_attributes(self, store, readonly=False):
        synchronizer = ThreadSynchronizer()
        return SynchronizedAttributes(store, synchronizer, readonly=readonly)


class TestProcessSynchronizedAttributes(TestAttributes):

    def init_attributes(self, store, readonly=False):
        sync_path = mkdtemp()
        atexit.register(shutil.rmtree, sync_path)
        synchronizer = ProcessSynchronizer(sync_path)
        return SynchronizedAttributes(store, synchronizer, readonly=readonly)


class TestThreadSynchronizedArray(TestArray):

    def create_array(self, store=None, readonly=False, **kwargs):
        if store is None:
            store = dict()
        init_store(store, **kwargs)
        return SynchronizedArray(store, synchronizer=ThreadSynchronizer(),
                                 readonly=readonly)


class TestProcessSynchronizedArray(TestArray):

    def create_array(self, store=None, readonly=False, **kwargs):
        if store is None:
            store = dict()
        init_store(store, **kwargs)
        sync_path = mkdtemp()
        atexit.register(shutil.rmtree, sync_path)
        synchronizer = ProcessSynchronizer(sync_path)
        return SynchronizedArray(store, synchronizer=synchronizer,
                                 readonly=readonly)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, division


import numpy as np


from zarr.compat import integer_types


def normalize_shape(shape):
    """Convenience function to normalize the `shape` argument."""

    # handle 1D convenience form
    if isinstance(shape, integer_types):
        shape = (int(shape),)

    # normalize
    shape = tuple(int(s) for s in shape)
    return shape


def normalize_chunks(chunks, shape):
    """Convenience function to normalize the `chunks` argument for an array
    with the given `shape`."""

    # N.B., expect shape already normalized

    # handle 1D convenience form
    if isinstance(chunks, integer_types):
        chunks = (int(chunks),)

    # handle bad dimensionality
    if len(chunks) > len(shape):
        raise ValueError('too many dimensions in chunks')

    # handle underspecified chunks
    if len(chunks) < len(shape):
        # assume chunks across remaining dimensions
        chunks += shape[len(chunks):]

    # handle None in chunks
    chunks = tuple(s if c is None else int(c)
                   for s, c in zip(shape, chunks))

    return chunks


# noinspection PyTypeChecker
def is_total_slice(item, shape):
    """Determine whether `item` specifies a complete slice of array with the
    given `shape`. Used to optimise __setitem__ operations on the Chunk
    class."""

    # N.B., assume shape is normalized

    if item == Ellipsis:
        return True
    if item == slice(None):
        return True
    if isinstance(item, slice):
        item = item,
    if isinstance(item, tuple):
        return all(
            (isinstance(s, slice) and
                ((s == slice(None)) or
                 ((s.stop - s.start == l) and (s.step in [1, None]))))
            for s, l in zip(item, shape)
        )
    else:
        raise TypeError('expected slice or tuple of slices, found %r' % item)


def normalize_axis_selection(item, l):
    """Convenience function to normalize a selection within a single axis
    of size `l`."""

    if isinstance(item, int):
        if item < 0:
            # handle wraparound
            item = l + item
        if item > (l - 1) or item < 0:
            raise IndexError('index out of bounds: %s' % item)
        return item

    elif isinstance(item, slice):
        if item.step is not None and item.step != 1:
            raise NotImplementedError('slice with step not supported')
        start = 0 if item.start is None else item.start
        stop = l if item.stop is None else item.stop
        if start < 0:
            start = l + start
        if stop < 0:
            stop = l + stop
        if start < 0 or stop < 0:
            raise IndexError('index out of bounds: %s, %s' % (start, stop))
        if start >= l:
            raise IndexError('index out of bounds: %s, %s' % (start, stop))
        if stop > l:
            stop = l
        if stop < start:
            raise IndexError('index out of bounds: %s, %s' % (start, stop))
        return slice(start, stop)

    else:
        raise TypeError('expected integer or slice, found: %r' % item)


# noinspection PyTypeChecker
def normalize_array_selection(item, shape):
    """Convenience function to normalize a selection within an array with
    the given `shape`."""

    # normalize item
    if isinstance(item, integer_types):
        item = (int(item),)
    elif isinstance(item, slice):
        item = (item,)
    elif item == Ellipsis:
        item = (slice(None),)

    # handle tuple of indices/slices
    if isinstance(item, tuple):

        # determine start and stop indices for all axes
        selection = tuple(normalize_axis_selection(i, l)
                          for i, l in zip(item, shape))

        # fill out selection if not completely specified
        if len(selection) < len(shape):
            selection += tuple(slice(0, l) for l in shape[len(selection):])

        return selection

    else:
        raise TypeError('expected indices or slice, found: %r' % item)


def get_chunk_range(selection, chunks):
    """Convenience function to get a range over all chunk indices,
    for iterating over chunks."""
    chunk_range = [range(s.start//l, int(np.ceil(s.stop/l)))
                   if isinstance(s, slice)
                   else range(s//l, (s//l)+1)
                   for s, l in zip(selection, chunks)]
    return chunk_range


def normalize_resize_args(old_shape, *args):

    # normalize new shape argument
    if len(args) == 1:
        new_shape = args[0]
    else:
        new_shape = args
    if isinstance(new_shape, int):
        new_shape = (new_shape,)
    else:
        new_shape = tuple(new_shape)
    if len(new_shape) != len(old_shape):
        raise ValueError('new shape must have same number of dimensions')

    # handle None in new_shape
    new_shape = tuple(s if n is None else int(n)
                      for s, n in zip(old_shape, new_shape))

    return new_shape


def human_readable_size(size):
    if size < 2**10:
        return '%s' % size
    elif size < 2**20:
        return '%.1fK' % (size / float(2**10))
    elif size < 2**30:
        return '%.1fM' % (size / float(2**20))
    elif size < 2**40:
        return '%.1fG' % (size / float(2**30))
    elif size < 2**50:
        return '%.1fT' % (size / float(2**40))
    else:
        return '%.1fP' % (size / float(2**50))


def normalize_order(order):
    order = str(order).upper()
    if order not in ['C', 'F']:
        raise ValueError("order must be either 'C' or 'F', found: %r" % order)
    return order



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, division
from glob import glob
import os
from setuptools import setup, Extension
import cpuinfo
import sys
from distutils.command.build_ext import build_ext
from distutils.errors import CCompilerError, DistutilsExecError, \
    DistutilsPlatformError


PY2 = sys.version_info[0] == 2


try:
    from Cython.Build import cythonize
except ImportError:
    have_cython = False
else:
    have_cython = True


def blosc_extension():
    print('[zarr] Setting up Blosc extension')

    # setup blosc extension
    blosc_sources = []
    extra_compile_args = []
    include_dirs = []
    define_macros = []

    # generic setup
    blosc_sources += [f for f in glob('c-blosc/blosc/*.c')
                      if 'avx2' not in f and 'sse2' not in f]
    blosc_sources += glob('c-blosc/internal-complibs/lz4*/*.c')
    blosc_sources += glob('c-blosc/internal-complibs/snappy*/*.cc')
    blosc_sources += glob('c-blosc/internal-complibs/zlib*/*.c')
    include_dirs += [os.path.join('c-blosc', 'blosc')]
    include_dirs += glob('c-blosc/internal-complibs/*')
    define_macros += [('HAVE_LZ4', 1), ('HAVE_SNAPPY', 1), ('HAVE_ZLIB', 1)]
    # define_macros += [('CYTHON_TRACE', '1')]

    # determine CPU support for SSE2 and AVX2
    cpu_info = cpuinfo.get_cpu_info()

    # SSE2
    if 'sse2' in cpu_info['flags']:
        print('[zarr] SSE2 detected')
        extra_compile_args.append('-DSHUFFLE_SSE2_ENABLED')
        blosc_sources += [f for f in glob('c-blosc/blosc/*.c') if 'sse2' in f]
        if os.name == 'posix':
            extra_compile_args.append('-msse2')
        elif os.name == 'nt':
            define_macros += [('__SSE2__', 1)]

    # AVX2
    if 'avx2' in cpu_info['flags']:
        print('[zarr] AVX2 detected')
        extra_compile_args.append('-DSHUFFLE_AVX2_ENABLED')
        blosc_sources += [f for f in glob('c-blosc/blosc/*.c') if 'avx2' in f]
        if os.name == 'posix':
            extra_compile_args.append('-mavx2')
        elif os.name == 'nt':
            define_macros += [('__AVX2__', 1)]

    # workaround lack of support for "inline" in MSVC when building for Python
    # 2.7 64-bit
    if PY2 and os.name == 'nt':
        extra_compile_args.append('-Dinline=__inline')

    if have_cython:
        sources = ['zarr/blosc.pyx']
    else:
        sources = ['zarr/blosc.c']

    # define extension module
    extensions = [
        Extension('zarr.blosc',
                  sources=sources + blosc_sources,
                  include_dirs=include_dirs,
                  define_macros=define_macros,
                  extra_compile_args=extra_compile_args,
                  ),
    ]

    if have_cython:
        extensions = cythonize(extensions)

    return extensions


if sys.platform == 'win32':
    ext_errors = (CCompilerError, DistutilsExecError, DistutilsPlatformError,
                  IOError, ValueError)
else:
    ext_errors = (CCompilerError, DistutilsExecError, DistutilsPlatformError)


class BuildFailed(Exception):
    pass


class ve_build_ext(build_ext):
    # This class allows C extension building to fail.

    def run(self):
        try:
            build_ext.run(self)
        except DistutilsPlatformError as e:
            print('[zarr] ERROR', e, file=sys.stderr)
            raise BuildFailed()

    def build_extension(self, ext):
        try:
            build_ext.build_extension(self, ext)
        except ext_errors as e:
            print('[zarr] ERROR', e, file=sys.stderr)
            raise BuildFailed()


DESCRIPTION = 'A minimal implementation of chunked, compressed, ' \
              'N-dimensional arrays for Python.'

with open('README.rst') as f:
    LONG_DESCRIPTION = f.read()


def run_setup(with_extensions):

    if with_extensions:
        ext_modules = blosc_extension()
        cmdclass = dict(build_ext=ve_build_ext)
    else:
        ext_modules = []
        cmdclass = dict()

    setup(
        name='zarr',
        description=DESCRIPTION,
        long_description=LONG_DESCRIPTION,
        use_scm_version={
            'version_scheme': 'guess-next-dev',
            'local_scheme': 'dirty-tag',
            'write_to': 'zarr/version.py'
        },
        setup_requires=[
            'setuptools>18.0',
            'setuptools-scm>1.5.4'
        ],
        install_requires=[
            'numpy>=1.7',
            'fasteners',
        ],
        ext_modules=ext_modules,
        cmdclass=cmdclass,
        package_dir={'': '.'},
        packages=['zarr', 'zarr.tests'],
        classifiers=[
            'Development Status :: 4 - Beta',
            'Intended Audience :: Developers',
            'Intended Audience :: Information Technology',
            'Intended Audience :: Science/Research',
            'License :: OSI Approved :: MIT License',
            'Programming Language :: Python',
            'Topic :: Software Development :: Libraries :: Python Modules',
            'Operating System :: Unix',
            'Programming Language :: Python :: 2',
            'Programming Language :: Python :: 2.7',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.4',
            'Programming Language :: Python :: 3.5',
        ],
        author='Alistair Miles',
        author_email='alimanfoo@googlemail.com',
        maintainer='Alistair Miles',
        maintainer_email='alimanfoo@googlemail.com',
        url='https://github.com/alimanfoo/zarr',
        license='MIT',
    )


if __name__ == '__main__':
    is_pypy = hasattr(sys, 'pypy_translation_info')

    try:
        run_setup(not is_pypy)
    except BuildFailed:
        print('[zarr]' + ('*' * 75), file=sys.stderr)
        print('[zarr] WARNING compilation of the Blosc C extension failed.',
              file=sys.stderr)
        print('[zarr] Retrying installation without C extensions...',
              file=sys.stderr)
        print('[zarr]' + ('*' * 75), file=sys.stderr)
        run_setup(False)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, division
import zlib
import bz2


import numpy as np


from zarr.compat import text_type


registry = dict()


class ZlibCompressor(object):
    """Provides compression using zlib via the Python standard library.
    Registered under the name 'zlib'.

    Parameters
    ----------
    compression_opts : int
        An integer between 0 and 9 inclusive specifying the compression level.

    Examples
    --------
    >>> import zarr
    >>> z = zarr.zeros((10000, 10000), chunks=(1000, 1000), dtype='i4',
    ...                compression='zlib', compression_opts=1)

    """

    canonical_name = 'zlib'
    default_level = 1

    def __init__(self, compression_opts):
        # at this point we expect compression_opts to be fully normalized
        self.level = compression_opts

    @classmethod
    def normalize_opts(cls, compression_opts):
        """Convenience function to normalize compression options."""
        if compression_opts is None:
            level = cls.default_level
        else:
            level = int(compression_opts)
        if level < 0 or level > 9:
            raise ValueError('invalid compression level: %s' % level)
        return level

    # noinspection PyMethodMayBeStatic
    def decompress(self, cdata, array):
        data = zlib.decompress(cdata)
        src = np.frombuffer(data, dtype=array.dtype).reshape(array.shape)
        np.copyto(array, src)

    def compress(self, array):
        data = array.tobytes()
        return zlib.compress(data, self.level)


registry[ZlibCompressor.canonical_name] = ZlibCompressor
registry['gzip'] = ZlibCompressor  # alias
default_compression = ZlibCompressor.canonical_name


try:
    from zarr import blosc
except ImportError:  # pragma: no cover
    pass
else:

    class BloscCompressor(object):
        """Provides compression using the blosc meta-compressor. Registered
        under the name 'blosc'.

        Parameters
        ----------
        compression_opts : dict
            A dictionary with keys 'cname', 'clevel' and 'shuffle'. The value
            of the 'cname' key should be a string naming one of the compression
            algorithms available within blosc, e.g., 'blosclz', 'lz4', 'zlib'
            or 'snappy'. The value of the 'clevel' key should be an integer
            between 0 and 9 specifying the compression level. The value of the
            'shuffle' key should be 0 (no shuffle), 1 (byte shuffle) or 2 (bit
            shuffle).

        Examples
        --------
        >>> import zarr
        >>> z = zarr.zeros((10000, 10000), chunks=(1000, 1000), dtype='i4',
        ...                compression='blosc',
        ...                compression_opts=dict(cname='lz4', clevel=3, shuffle=2))

        """  # flake8: noqa

        canonical_name = 'blosc'
        default_cname = 'blosclz'
        default_clevel = 5
        default_shuffle = 1

        def __init__(self, compression_opts):
            # at this point we expect compression_opts to be fully normalized
            cname = compression_opts['cname']
            if isinstance(cname, text_type):
                cname = cname.encode('ascii')
            self.cname = cname
            self.clevel = compression_opts['clevel']
            self.shuffle = compression_opts['shuffle']

        @classmethod
        def normalize_opts(cls, compression_opts):
            """Convenience function to normalize compression options."""

            if compression_opts is None:
                compression_opts = dict()
            cname = compression_opts.get('cname', None)
            clevel = compression_opts.get('clevel', None)
            shuffle = compression_opts.get('shuffle', None)

            # determine internal compression library
            cname = cname if cname is not None else cls.default_cname

            # check internal compressor is available
            if blosc.compname_to_compcode(cname) < 0:
                raise ValueError('blosc internal compressor not available: %s'
                                 % cname)

            # determine compression level
            clevel = clevel if clevel is not None else cls.default_clevel
            clevel = int(clevel)
            if clevel < 0 or clevel > 9:
                raise ValueError('invalid compression level: %s' % clevel)

            # determine shuffle filter
            shuffle = shuffle if shuffle is not None else cls.default_shuffle
            shuffle = int(shuffle)
            if shuffle not in [0, 1, 2]:
                raise ValueError('invalid shuffle: %s' % shuffle)

            # construct normalised options
            compression_opts = dict(
                cname=cname, clevel=clevel, shuffle=shuffle
            )
            return compression_opts

        # noinspection PyMethodMayBeStatic
        def decompress(self, cdata, array):
            blosc.decompress(cdata, array)

        def compress(self, array):
            return blosc.compress(array, self.cname, self.clevel, self.shuffle)

    registry[BloscCompressor.canonical_name] = BloscCompressor
    default_compression = BloscCompressor.canonical_name


class BZ2Compressor(object):
    """Provides compression using BZ2 via the Python standard library.
    Registered under the name 'bz2'.

    Parameters
    ----------
    compression_opts : int
        An integer between 1 and 9 inclusive specifying the compression level.

    Examples
    --------
    >>> import zarr
    >>> z = zarr.zeros((10000, 10000), chunks=(1000, 1000), dtype='i4',
    ...                compression='bz2', compression_opts=1)
    """

    canonical_name = 'bz2'
    default_level = 1

    def __init__(self, compression_opts):
        # at this point we expect compression_opts to be fully normalized
        self.level = compression_opts

    @classmethod
    def normalize_opts(cls, compression_opts):
        """Convenience function to normalize compression options."""
        if compression_opts is None:
            level = cls.default_level
        else:
            level = int(compression_opts)
        if level < 1 or level > 9:
            raise ValueError('invalid compression level: %s' % level)
        return level

    # noinspection PyMethodMayBeStatic
    def decompress(self, cdata, array):
        data = bz2.decompress(cdata)
        src = np.frombuffer(data, dtype=array.dtype).reshape(array.shape)
        np.copyto(array, src)

    def compress(self, array):
        data = array.tobytes()
        return bz2.compress(data, self.level)


registry[BZ2Compressor.canonical_name] = BZ2Compressor


try:
    import lzma
except ImportError:  # pragma: no cover
    pass
else:

    class LZMACompressor(object):
        """Provides compression using lzma via the Python standard library
        (only available under Python 3). Registered under the name 'lzma'.

        Parameters
        ----------
        compression_opts : dict
            A dictionary with keys 'format', 'check', 'preset' and 'filters'.
            The value of the 'format' key should be an integer specifying one
            of the lzma format codes, e.g., ``lzma.FORMAT_XZ``. The value of
            the 'check' key should be an integer specifying one of the lzma
            check codes, e.g., ``lzma.CHECK_NONE``. The value of the 'preset'
            key should be an integer between 0 and 9 inclusive, specifying the
            compression level. The value of the 'filters' key should be a list
            of dictionaries specifying compression filters. If filters are
            provided, 'preset' must be None.

        Examples
        --------
        Simple usage::

            >>> import zarr
            >>> z = zarr.zeros((10000, 10000), chunks=(1000, 1000), dtype='i4',
            ...                compression='lzma',
            ...                compression_opts=dict(preset=1))

        Custom filter pipeline::

            >>> import lzma
            >>> filters = [dict(id=lzma.FILTER_DELTA, dist=4),
            ...            dict(id=lzma.FILTER_LZMA2, preset=1)]
            >>> z = zarr.zeros((10000, 10000), chunks=(1000, 1000), dtype='i4',
            ...                compression='lzma',
            ...                compression_opts=dict(filters=filters))

        """

        canonical_name = 'lzma'
        default_format = lzma.FORMAT_XZ
        default_check = lzma.CHECK_NONE
        default_preset = 1
        default_filters = None

        def __init__(self, compression_opts):
            # at this point we expect compression_opts to be fully normalized
            self.format = compression_opts['format']
            self.check = compression_opts['check']
            self.preset = compression_opts['preset']
            self.filters = compression_opts['filters']

        @classmethod
        def normalize_opts(cls, compression_opts):
            """Convenience function to normalize compression options."""

            if compression_opts is None:
                compression_opts = dict()

            format = compression_opts.get('format', None)
            check = compression_opts.get('check', None)
            preset = compression_opts.get('preset', None)
            filters = compression_opts.get('filters', None)

            # normalize format
            if format is None:
                format = cls.default_format
            if format not in [lzma.FORMAT_XZ, lzma.FORMAT_ALONE,
                              lzma.FORMAT_RAW]:
                raise ValueError('invalid format: %s' % format)

            # normalize check
            if check is None:
                check = cls.default_check
            if check not in [lzma.CHECK_NONE, lzma.CHECK_CRC32,
                             lzma.CHECK_CRC64, lzma.CHECK_SHA256]:
                raise ValueError('invalid check: %s' % check)

            # normalize preset
            if preset is None:
                preset = cls.default_preset
            if preset < 0 or preset > 9:
                raise ValueError('invalid preset: %s' % preset)

            # handle filters
            if filters:
                # cannot specify both preset and filters
                preset = None

            # construct normalized options
            compression_opts = dict(
                format=format, check=check, preset=preset, filters=filters
            )
            return compression_opts

        # noinspection PyMethodMayBeStatic
        def decompress(self, cdata, array):
            if self.format == lzma.FORMAT_RAW:
                # filters needed
                filters = self.filters
            else:
                # filters should not be specified
                filters = None
            data = lzma.decompress(cdata, format=self.format, filters=filters)
            src = np.frombuffer(data, dtype=array.dtype).reshape(array.shape)
            np.copyto(array, src)

        def compress(self, array):
            data = array.tobytes()
            return lzma.compress(data, format=self.format, check=self.check,
                                 preset=self.preset, filters=self.filters)

    registry[LZMACompressor.canonical_name] = LZMACompressor


class NoCompressor(object):
    """No compression, i.e., pass bytes through. Registered under the
    name 'none'.

    Examples
    --------
    >>> import zarr
    >>> z = zarr.zeros((10000, 10000), chunks=(1000, 1000), dtype='i4',
    ...                compression='none')

    """

    canonical_name = 'none'

    def __init__(self, compression_opts):
        pass

    @classmethod
    def normalize_opts(cls, compression_opts):
        if compression_opts is not None:
            raise ValueError('no compression options supported')
        return None

    # noinspection PyMethodMayBeStatic
    def decompress(self, cdata, array):
        src = np.frombuffer(cdata, dtype=array.dtype).reshape(array.shape)
        np.copyto(array, src)

    # noinspection PyMethodMayBeStatic
    def compress(self, array):
        data = array.tobytes()
        return data


registry[NoCompressor.canonical_name] = NoCompressor
registry[None] = NoCompressor  # alias


def get_compressor_cls(compression):
    if compression == 'default':
        compression = default_compression
    try:
        return registry[compression]
    except KeyError:
        raise ValueError('compressor not available: %r' % compression)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, division
import unittest
from tempfile import mkdtemp
import atexit
import shutil
import pickle


import numpy as np
from numpy.testing import assert_array_equal
from nose.tools import eq_ as eq, assert_is_instance, \
    assert_raises, assert_true, assert_false
import zict
from zarr.storage import DirectoryStore, init_store


from zarr.core import Array
from zarr.errors import ReadOnlyError


def test_array_init():

    store = dict()  # store not initialised
    with assert_raises(ValueError):
        Array(store)


def test_nbytes_stored():

    store = dict()
    init_store(store, shape=1000, chunks=100)
    z = Array(store)
    eq(sum(len(v) for v in store.values()), z.nbytes_stored)
    z[:] = 42
    eq(sum(len(v) for v in store.values()), z.nbytes_stored)

    # store supporting size determination
    path = mkdtemp()
    atexit.register(shutil.rmtree, path)
    store = DirectoryStore(path)
    init_store(store, shape=1000, chunks=100)
    z = Array(store)
    eq(sum(len(v) for v in store.values()), z.nbytes_stored)
    z[:] = 42
    eq(sum(len(v) for v in store.values()), z.nbytes_stored)

    # custom store, doesn't support size determination
    store = zict.Zip('test.zip', mode='w')
    init_store(store, shape=1000, chunks=100, compression='zlib',
               compression_opts=1)
    z = Array(store)
    eq(-1, z.nbytes_stored)
    z[:] = 42
    eq(-1, z.nbytes_stored)


class TestArray(unittest.TestCase):

    def create_array(self, store=None, readonly=False, **kwargs):
        if store is None:
            store = dict()
        init_store(store, **kwargs)
        return Array(store, readonly=readonly)

    def test_array_1d(self):

        a = np.arange(1050)
        z = self.create_array(shape=a.shape, chunks=100, dtype=a.dtype)

        # check properties
        eq(len(a), len(z))
        eq(a.shape, z.shape)
        eq(a.dtype, z.dtype)
        eq((100,), z.chunks)
        eq(a.nbytes, z.nbytes)
        eq(sum(len(v) for v in z.store.values()), z.nbytes_stored)
        eq(0, z.initialized)
        eq((11,), z.cdata_shape)

        # check empty
        b = z[:]
        assert_is_instance(b, np.ndarray)
        eq(a.shape, b.shape)
        eq(a.dtype, b.dtype)

        # check attributes
        z.attrs['foo'] = 'bar'
        eq('bar', z.attrs['foo'])

        # set data
        z[:] = a

        # check properties
        eq(a.nbytes, z.nbytes)
        eq(sum(len(v) for v in z.store.values()), z.nbytes_stored)
        eq(11, z.initialized)

        # check slicing
        assert_array_equal(a, np.array(z))
        assert_array_equal(a, z[:])
        assert_array_equal(a, z[...])
        # noinspection PyTypeChecker
        assert_array_equal(a, z[slice(None)])
        assert_array_equal(a[:10], z[:10])
        assert_array_equal(a[10:20], z[10:20])
        assert_array_equal(a[-10:], z[-10:])
        # ...across chunk boundaries...
        assert_array_equal(a[:110], z[:110])
        assert_array_equal(a[190:310], z[190:310])
        assert_array_equal(a[-110:], z[-110:])
        # single item
        eq(a[0], z[0])
        eq(a[-1], z[-1])

        # check partial assignment
        b = np.arange(1e5, 2e5)
        z[190:310] = b[190:310]
        assert_array_equal(a[:190], z[:190])
        assert_array_equal(b[190:310], z[190:310])
        assert_array_equal(a[310:], z[310:])

    def test_array_1d_fill_value(self):

        for fill_value in -1, 0, 1, 10:

            a = np.arange(1050)
            f = np.empty_like(a)
            f.fill(fill_value)
            z = self.create_array(shape=a.shape, chunks=100, dtype=a.dtype,
                                  fill_value=fill_value)
            z[190:310] = a[190:310]

            assert_array_equal(f[:190], z[:190])
            assert_array_equal(a[190:310], z[190:310])
            assert_array_equal(f[310:], z[310:])

    def test_array_1d_set_scalar(self):

        # setup
        a = np.empty(100)
        z = self.create_array(shape=a.shape, chunks=10, dtype=a.dtype)
        z[:] = a
        assert_array_equal(a, z[:])

        for value in -1, 0, 1, 10:
            a[15:35] = value
            z[15:35] = value
            assert_array_equal(a, z[:])
            a[:] = value
            z[:] = value
            assert_array_equal(a, z[:])

    def test_array_2d(self):

        a = np.arange(10000).reshape((1000, 10))
        z = self.create_array(shape=a.shape, chunks=(100, 2), dtype=a.dtype)

        # check properties
        eq(len(a), len(z))
        eq(a.shape, z.shape)
        eq(a.dtype, z.dtype)
        eq((100, 2), z.chunks)
        eq(sum(len(v) for v in z.store.values()), z.nbytes_stored)
        eq(0, z.initialized)
        eq((10, 5), z.cdata_shape)

        # set data
        z[:] = a

        # check properties
        eq(a.nbytes, z.nbytes)
        eq(sum(len(v) for v in z.store.values()), z.nbytes_stored)
        eq(50, z.initialized)

        # check slicing
        assert_array_equal(a, np.array(z))
        assert_array_equal(a, z[:])
        assert_array_equal(a, z[...])
        # noinspection PyTypeChecker
        assert_array_equal(a, z[slice(None)])
        assert_array_equal(a[:10], z[:10])
        assert_array_equal(a[10:20], z[10:20])
        assert_array_equal(a[-10:], z[-10:])
        assert_array_equal(a[:, :2], z[:, :2])
        assert_array_equal(a[:, 2:4], z[:, 2:4])
        assert_array_equal(a[:, -2:], z[:, -2:])
        assert_array_equal(a[:10, :2], z[:10, :2])
        assert_array_equal(a[10:20, 2:4], z[10:20, 2:4])
        assert_array_equal(a[-10:, -2:], z[-10:, -2:])
        # ...across chunk boundaries...
        assert_array_equal(a[:110], z[:110])
        assert_array_equal(a[190:310], z[190:310])
        assert_array_equal(a[-110:], z[-110:])
        assert_array_equal(a[:, :3], z[:, :3])
        assert_array_equal(a[:, 3:7], z[:, 3:7])
        assert_array_equal(a[:, -3:], z[:, -3:])
        assert_array_equal(a[:110, :3], z[:110, :3])
        assert_array_equal(a[190:310, 3:7], z[190:310, 3:7])
        assert_array_equal(a[-110:, -3:], z[-110:, -3:])
        # single item
        assert_array_equal(a[0], z[0])
        assert_array_equal(a[-1], z[-1])
        eq(a[0, 0], z[0, 0])
        eq(a[-1, -1], z[-1, -1])

        # check partial assignment
        b = np.arange(10000, 20000).reshape((1000, 10))
        z[190:310, 3:7] = b[190:310, 3:7]
        assert_array_equal(a[:190], z[:190])
        assert_array_equal(a[:, :3], z[:, :3])
        assert_array_equal(b[190:310, 3:7], z[190:310, 3:7])
        assert_array_equal(a[310:], z[310:])
        assert_array_equal(a[:, 7:], z[:, 7:])

    def test_array_2d_partial(self):

        z = self.create_array(shape=(1000, 10), chunks=(100, 2), dtype='i4',
                              fill_value=0)

        # check partial assignment, single row
        c = np.arange(z.shape[1])
        z[0, :] = c
        with assert_raises(ValueError):
            # N.B., NumPy allows this, but we'll be strict for now
            z[2:3] = c
        with assert_raises(ValueError):
            # N.B., NumPy allows this, but we'll be strict for now
            z[-1:] = c
        z[2:3] = c[None, :]
        z[-1:] = c[None, :]
        assert_array_equal(c, z[0, :])
        assert_array_equal(c, z[2, :])
        assert_array_equal(c, z[-1, :])

        # check partial assignment, single column
        d = np.arange(z.shape[0])
        z[:, 0] = d
        with assert_raises(ValueError):
            z[:, 2:3] = d
        with assert_raises(ValueError):
            z[:, -1:] = d
        z[:, 2:3] = d[:, None]
        z[:, -1:] = d[:, None]
        assert_array_equal(d, z[:, 0])
        assert_array_equal(d, z[:, 2])
        assert_array_equal(d, z[:, -1])

        # check single item assignment
        z[0, 0] = -1
        z[2, 2] = -1
        z[-1, -1] = -1
        eq(-1, z[0, 0])
        eq(-1, z[2, 2])
        eq(-1, z[-1, -1])

    def test_array_order(self):

        # 1D
        a = np.arange(1050)
        for order in 'C', 'F':
            z = self.create_array(shape=a.shape, chunks=100, dtype=a.dtype,
                                  order=order)
            eq(order, z.order)
            if order == 'F':
                assert_true(z[:].flags.f_contiguous)
            else:
                assert_true(z[:].flags.c_contiguous)
            z[:] = a
            assert_array_equal(a, z[:])

        # 2D
        a = np.arange(10000).reshape((100, 100))
        for order in 'C', 'F':
            z = self.create_array(shape=a.shape, chunks=(10, 10),
                                  dtype=a.dtype, order=order)
            eq(order, z.order)
            if order == 'F':
                assert_true(z[:].flags.f_contiguous)
            else:
                assert_true(z[:].flags.c_contiguous)
            z[:] = a
            assert_array_equal(a, z[:])

    def test_resize_1d(self):

        z = self.create_array(shape=105, chunks=10, dtype='i4', fill_value=0)
        a = np.arange(105, dtype='i4')
        z[:] = a
        eq((105,), z.shape)
        eq((105,), z[:].shape)
        eq(np.dtype('i4'), z.dtype)
        eq(np.dtype('i4'), z[:].dtype)
        eq((10,), z.chunks)
        assert_array_equal(a, z[:])

        z.resize(205)
        eq((205,), z.shape)
        eq((205,), z[:].shape)
        eq(np.dtype('i4'), z.dtype)
        eq(np.dtype('i4'), z[:].dtype)
        eq((10,), z.chunks)
        assert_array_equal(a, z[:105])
        assert_array_equal(np.zeros(100, dtype='i4'), z[105:])

        z.resize(55)
        eq((55,), z.shape)
        eq((55,), z[:].shape)
        eq(np.dtype('i4'), z.dtype)
        eq(np.dtype('i4'), z[:].dtype)
        eq((10,), z.chunks)
        assert_array_equal(a[:55], z[:])

    def test_resize_2d(self):

        z = self.create_array(shape=(105, 105), chunks=(10, 10), dtype='i4',
                              fill_value=0)
        a = np.arange(105*105, dtype='i4').reshape((105, 105))
        z[:] = a
        eq((105, 105), z.shape)
        eq((105, 105), z[:].shape)
        eq(np.dtype('i4'), z.dtype)
        eq(np.dtype('i4'), z[:].dtype)
        eq((10, 10), z.chunks)
        assert_array_equal(a, z[:])

        z.resize((205, 205))
        eq((205, 205), z.shape)
        eq((205, 205), z[:].shape)
        eq(np.dtype('i4'), z.dtype)
        eq(np.dtype('i4'), z[:].dtype)
        eq((10, 10), z.chunks)
        assert_array_equal(a, z[:105, :105])
        assert_array_equal(np.zeros((100, 205), dtype='i4'), z[105:, :])
        assert_array_equal(np.zeros((205, 100), dtype='i4'), z[:, 105:])

        z.resize((55, 55))
        eq((55, 55), z.shape)
        eq((55, 55), z[:].shape)
        eq(np.dtype('i4'), z.dtype)
        eq(np.dtype('i4'), z[:].dtype)
        eq((10, 10), z.chunks)
        assert_array_equal(a[:55, :55], z[:])

        z.resize((55, 1))
        eq((55, 1), z.shape)
        eq((55, 1), z[:].shape)
        eq(np.dtype('i4'), z.dtype)
        eq(np.dtype('i4'), z[:].dtype)
        eq((10, 10), z.chunks)
        assert_array_equal(a[:55, :1], z[:])

    def test_append_1d(self):

        a = np.arange(105)
        z = self.create_array(shape=a.shape, chunks=10, dtype=a.dtype)
        z[:] = a
        eq(a.shape, z.shape)
        eq(a.dtype, z.dtype)
        eq((10,), z.chunks)
        assert_array_equal(a, z[:])

        b = np.arange(105, 205)
        e = np.append(a, b)
        z.append(b)
        eq(e.shape, z.shape)
        eq(e.dtype, z.dtype)
        eq((10,), z.chunks)
        assert_array_equal(e, z[:])

        # check append handles array-like
        c = [1, 2, 3]
        f = np.append(e, c)
        z.append(c)
        eq(f.shape, z.shape)
        eq(f.dtype, z.dtype)
        eq((10,), z.chunks)
        assert_array_equal(f, z[:])

    def test_append_2d(self):

        a = np.arange(105*105, dtype='i4').reshape((105, 105))
        z = self.create_array(shape=a.shape, chunks=(10, 10), dtype=a.dtype)
        z[:] = a
        eq(a.shape, z.shape)
        eq(a.dtype, z.dtype)
        eq((10, 10), z.chunks)
        assert_array_equal(a, z[:])

        b = np.arange(105*105, 2*105*105, dtype='i4').reshape((105, 105))
        e = np.append(a, b, axis=0)
        z.append(b)
        eq(e.shape, z.shape)
        eq(e.dtype, z.dtype)
        eq((10, 10), z.chunks)
        assert_array_equal(e, z[:])

    def test_append_2d_axis(self):

        a = np.arange(105*105, dtype='i4').reshape((105, 105))
        z = self.create_array(shape=a.shape, chunks=(10, 10), dtype=a.dtype)
        z[:] = a
        eq(a.shape, z.shape)
        eq(a.dtype, z.dtype)
        eq((10, 10), z.chunks)
        assert_array_equal(a, z[:])

        b = np.arange(105*105, 2*105*105, dtype='i4').reshape((105, 105))
        e = np.append(a, b, axis=1)
        z.append(b, axis=1)
        eq(e.shape, z.shape)
        eq(e.dtype, z.dtype)
        eq((10, 10), z.chunks)
        assert_array_equal(e, z[:])

    def test_append_bad_shape(self):
        a = np.arange(100)
        z = self.create_array(shape=a.shape, chunks=10, dtype=a.dtype)
        z[:] = a
        b = a.reshape(10, 10)
        with assert_raises(ValueError):
            z.append(b)

    def test_readonly(self):

        z = self.create_array(shape=1000, chunks=100)
        assert_false(z.readonly)

        z = self.create_array(shape=1000, chunks=100, readonly=True)
        assert_true(z.readonly)
        with assert_raises(ReadOnlyError):
            z[:] = 42
        with assert_raises(ReadOnlyError):
            z.resize(2000)
        with assert_raises(ReadOnlyError):
            z.append(np.arange(1000))

    def test_pickle(self):

        memory_store = dict()
        path = mkdtemp()
        atexit.register(shutil.rmtree, path)
        directory_store = DirectoryStore(path)

        for store in memory_store, directory_store:
            z = self.create_array(shape=1000, chunks=100, dtype=int,
                                  store=store)
            z[:] = np.random.randint(0, 1000, 1000)
            z2 = pickle.loads(pickle.dumps(z))
            eq(z.shape, z2.shape)
            eq(z.chunks, z2.chunks)
            eq(z.dtype, z2.dtype)
            eq(z.compression, z2.compression)
            eq(z.compression_opts, z2.compression_opts)
            eq(z.fill_value, z2.fill_value)
            assert_array_equal(z[:], z2[:])



# -*- coding: utf-8 -*-
# flake8: noqa
from __future__ import absolute_import, print_function, division
import sys


PY2 = sys.version_info[0] == 2


if PY2:  # pragma: no cover

    def itervalues(d, **kw):
        return d.itervalues(**kw)

    text_type = unicode
    binary_type = str
    integer_types = (int, long)

else:

    def itervalues(d, **kw):
        return iter(d.values(**kw))

    text_type = str
    binary_type = bytes
    integer_types = int,
